import getpass
import json
import os
import re
import sys
import time
import urllib.parse
import warnings
from pathlib import Path

import appdirs
import httpx

from .._version import __version__ as tiled_version
from ..utils import UNSET, DictView
from .auth import CannotRefreshAuthentication, TiledAuth, build_refresh_request
from .decoders import SUPPORTED_DECODERS
from .transport import Transport
from .utils import DEFAULT_TIMEOUT_PARAMS, MSGPACK_MIME_TYPE, handle_error

USER_AGENT = f"python-tiled/{tiled_version}"
API_KEY_AUTH_HEADER_PATTERN = re.compile(r"^Apikey (\w+)$")
PROMPT_FOR_REAUTHENTICATION = None


class Context:
    """
    Wrap an httpx.Client with an optional cache and authentication functionality.
    """

    def __init__(
        self,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=UNSET,
        timeout=None,
        verify=True,
        token_cache=None,
        app=None,
        raise_server_exceptions=True,
    ):
        # The uri is expected to reach the root API route.
        uri = httpx.URL(uri)
        headers = headers or {}
        # Define this here instead of at module scope so that the SUPPORTED_DECODERS
        # may be modified before Context instantiation.
        ACCEPT_ENCODING = ", ".join(
            [key for key in SUPPORTED_DECODERS.keys() if key != "identity"]
        )
        # Resolve this here, not at module scope, because the test suite
        # injects TILED_CACHE_DIR env var to use a temporary directory.
        TILED_CACHE_DIR = Path(
            os.getenv("TILED_CACHE_DIR", appdirs.user_cache_dir("tiled"))
        )
        headers.setdefault("accept-encoding", ACCEPT_ENCODING)
        # Set the User Agent to help the server fail informatively if the client
        # version is too old.
        headers.setdefault("user-agent", USER_AGENT)
        if token_cache is None:
            token_cache = TILED_CACHE_DIR / "tokens"

        # If ?api_key=... is present, move it from the query into a header.
        # The server would accept it in the query parameter, but using
        # a header is a little more secure (e.g. not logged).
        parsed_params = urllib.parse.parse_qs(uri.query.decode())
        api_key_list = parsed_params.pop("api_key", None)
        if api_key_list is not None:
            if api_key is not None:
                raise ValueError(
                    "api_key was provided as query parameter in URI and as keyword argument. Pick one."
                )
            if len(api_key_list) != 1:
                raise ValueError("Cannot handle two api_key query parameters")
            (api_key,) = api_key_list
        if api_key is None:
            # Check for an API key from the environment.
            api_key = os.getenv("TILED_API_KEY")
        # We will set the API key via the `api_key` property below,
        # after constructing the Client object.

        # FastAPI redirects /api -> /api/ so add it here to save a request.
        path = uri.path
        if not path.endswith("/"):
            path = f"{path}/"
        # Construct the uri *without* api_key param.
        # Drop any params/fragments.
        self.api_uri = httpx.URL(
            urllib.parse.urlunsplit((uri.scheme, uri.netloc.decode(), path, {}, ""))
        )
        if timeout is None:
            timeout = httpx.Timeout(**DEFAULT_TIMEOUT_PARAMS)
        if cache is UNSET:
            cache = None
        if app is None:
            client = httpx.Client(
                transport=Transport(cache=cache),
                verify=verify,
                timeout=timeout,
                follow_redirects=True,
            )
            # Do this in the setter to avoid being overwritten.
            client.headers = headers
        else:
            # Set up an ASGI client.
            # Because we have been handed an app, we can infer that
            # starlette is available.
            from starlette.testclient import TestClient

            # verify parameter is dropped, as there is no SSL in ASGI mode
            client = TestClient(
                app=app,
                raise_server_exceptions=raise_server_exceptions,
            )
            client.timeout = timeout
            client.headers = headers
            # Do this in the setter to avoid being overwritten.
            client.follow_redirects = True
            client._transport = Transport(transport=client._transport, cache=cache)
            client.__enter__()
            # The TestClient is meant to be used only as a context manager,
            # where the context starts and stops and the wrapped ASGI app.
            # We are abusing it slightly to enable interactive use of the
            # TestClient.
            if sys.version_info < (3, 9):
                import atexit

                atexit.register(client.__exit__)
            else:
                import threading

                # The threading module has its own (internal) atexit
                # mechanism that runs at thread shutdown, prior to the atexit
                # mechanism that runs at interpreter shutdown.
                # We need to intervene at that layer to close the portal, or else
                # we will wait forever for a thread run by the portal to join().
                threading._register_atexit(client.__exit__)

        self.http_client = client
        self._verify = verify
        self._cache = cache
        self._token_cache = Path(token_cache)

        # Make an initial "safe" request to:
        # (1) Get the server_info.
        # (2) Let the server set the CSRF cookie.
        # No authentication has been set up yet, so these requests will be unauthenticated.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        self.server_info = handle_error(
            self.http_client.get(
                self.api_uri,
                headers={
                    "Accept": MSGPACK_MIME_TYPE,
                    "Cache-Control": "no-cache, no-store",
                },
            )
        ).json()
        self.api_key = api_key  # property setter sets Authorization header
        self.admin = Admin(self)  # accessor for admin-related requests

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self.http_client.__exit__()

    def __getstate__(self):
        if getattr(self.http_client, "app", None):
            raise TypeError(
                "Cannot pickle a Tiled Context built around an ASGI. "
                "Only Tiled Context connected to remote servers can be pickled."
            )
        return (
            tiled_version,
            self.api_uri,
            self.http_client.headers,
            list(self.http_client.cookies.jar),
            self.http_client.timeout,
            self.http_client.auth,
            self._verify,
            self._token_cache,
            self.server_info,
            self.cache,
        )

    def __setstate__(self, state):
        (
            state_tiled_version,
            api_uri,
            headers,
            cookies_list,
            timeout,
            auth,
            verify,
            token_cache,
            server_info,
            cache,
        ) = state
        if state_tiled_version != tiled_version:
            raise TypeError(
                f"Cannot unpickle {type(self).__name__} from tiled version {state_tiled_version} "
                f"using tiled version {tiled_version}. Pickle should only be used to short-term "
                "transfer between identical versions of tiled."
            )
        self.api_uri = api_uri
        cookies = httpx.Cookies()
        for cookie in cookies_list:
            cookies.set(
                cookie.name, cookie.value, domain=cookie.domain, path=cookie.path
            )
        self.http_client = httpx.Client(
            verify=verify,
            cookies=cookies,
            timeout=timeout,
            headers=headers,
            follow_redirects=True,
            auth=auth,
        )
        self._token_cache = token_cache
        self._cache = cache
        self.server_info = server_info

    @classmethod
    def from_any_uri(
        cls,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=UNSET,
        timeout=None,
        verify=True,
        token_cache=None,
        app=None,
    ):
        """
        Accept a URI to a specific node.

        For example, given URI "https://example.com/api/v1//metadata/a/b/c"
        return a Context connected to "https://examples/api/v1" and the list
        ["a", "b", "c"].
        """
        uri = httpx.URL(uri)
        node_path_parts = []
        if "/metadata" in uri.path:
            api_path, _, node_path = uri.path.partition("/metadata")
            api_uri = uri.copy_with(path=api_path)
            node_path_parts.extend(
                [segment for segment in node_path.split("/") if segment]
            )
        # Below we handle the case where we are given *less* than the full URI
        # to the root endpoint. Here we are taking some care to plan for the case
        # where tiled is served at a sub-path, even though that is not yet supported
        # on the server side.
        elif "/api" not in uri.path:
            # Looks like we were given the root path (to the HTML landing page).
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/api/v1")
        elif "/v1" not in uri.path:
            # Looks like we were given the /api but no version.
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/v1")
        else:
            api_uri = uri
        context = cls(
            api_uri,
            headers=headers,
            api_key=api_key,
            cache=cache,
            timeout=timeout,
            verify=verify,
            token_cache=token_cache,
            app=app,
        )
        return context, node_path_parts

    @classmethod
    def from_app(
        cls,
        app,
        *,
        cache=UNSET,
        token_cache=None,
        headers=None,
        timeout=None,
        api_key=UNSET,
        raise_server_exceptions=True,
    ):
        """
        Construct a Context around a FastAPI app. Primarily for testing.
        """
        context = cls(
            uri="http://local-tiled-app/api/v1",
            headers=headers,
            api_key=api_key,
            cache=cache,
            timeout=timeout,
            token_cache=token_cache,
            app=app,
            raise_server_exceptions=raise_server_exceptions,
        )
        if (api_key is UNSET) and (
            not context.server_info["authentication"]["providers"]
        ):
            # Extract the API key from the app and set it.
            from ..server.settings import get_settings

            settings = app.dependency_overrides[get_settings]()
            api_key = settings.single_user_api_key or None
            context.api_key = api_key
        return context

    @property
    def tokens(self):
        "A view of the current access and refresh tokens."
        return DictView(self.http_client.auth.tokens)

    @property
    def api_key(self):
        # Extract from header to ensure that there is one "ground truth" here
        # and no possibility of state getting out of sync.
        header = self.http_client.headers.get("Authorization", "")
        match = API_KEY_AUTH_HEADER_PATTERN.match(header)
        if match is not None:
            return match.group(1)

    @api_key.setter
    def api_key(self, api_key):
        if api_key is None:
            if self.http_client.headers.get("Authorization", "").startswith("Apikey "):
                self.http_client.headers.pop("Authorization")
        else:
            self.http_client.headers["Authorization"] = f"Apikey {api_key}"

    @property
    def cache(self):
        return self.http_client._transport.cache

    @cache.setter
    def cache(self, cache):
        self.http_client._transport.cache = cache

    def which_api_key(self):
        """
        A 'who am I' for API keys
        """
        if not self.api_key:
            raise RuntimeError("Not API key is configured for the client.")
        return handle_error(
            self.http_client.get(
                self.server_info["authentication"]["links"]["apikey"],
                headers={"Accept": MSGPACK_MIME_TYPE},
            )
        ).json()

    def create_api_key(self, scopes=None, expires_in=None, note=None):
        """
        Generate a new API for the currently-authenticated user.

        Parameters
        ----------
        scopes : Optional[List[str]]
            Restrict the access available to the API key by listing specific scopes.
            By default, this will have the same access as the user.
        expires_in : Optional[int]
            Number of seconds until API key expires. If None,
            it will never expire or it will have the maximum lifetime
            allowed by the server.
        note : Optional[str]
            Description (for humans).
        """
        return handle_error(
            self.http_client.post(
                self.server_info["authentication"]["links"]["apikey"],
                headers={"Accept": MSGPACK_MIME_TYPE},
                json={"scopes": scopes, "expires_in": expires_in, "note": note},
            )
        ).json()

    def revoke_api_key(self, first_eight):
        handle_error(
            self.http_client.delete(
                self.server_info["authentication"]["links"]["apikey"],
                headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
                params={"first_eight": first_eight},
            )
        )

    @property
    def app(self):
        warnings.warn(
            "The 'app' accessor on Context is deprecated. Use Context.http_client.app.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.app

    @property
    def base_url(self):
        warnings.warn(
            "The 'base_url' accessor on Context is deprecated. Use Context.http_client.base_url.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.base_url

    @property
    def event_hooks(self):
        "httpx.Client event hooks. This is exposed for testing."
        warnings.warn(
            "The 'event_hooks' accessor on Context is deprecated. Use Context.http_client.event_hooks.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.event_hooks

    def authenticate(
        self,
        username=UNSET,
        provider=UNSET,
        prompt_for_reauthentication=UNSET,
        set_default=True,
    ):
        """
        See login. This is for programmatic use.
        """
        if prompt_for_reauthentication is UNSET:
            prompt_for_reauthentication = PROMPT_FOR_REAUTHENTICATION
        if prompt_for_reauthentication is None:
            prompt_for_reauthentication = _can_prompt()
        if (username is UNSET) and (provider is UNSET):
            default_identity = get_default_identity(self.api_uri)
            if default_identity is not None:
                username = default_identity["username"]
                provider = default_identity["provider"]
        if username is UNSET:
            username = None
        if provider is UNSET:
            provider = None
        providers = self.server_info["authentication"]["providers"]
        spec = _choose_identity_provider(providers, provider)
        provider = spec["provider"]
        if self.api_key is not None:
            # Check that API key authenticates us as this user,
            # and then either return or raise.
            identities = self.whoami()["identities"]
            for identity in identities:
                if (identity["provider"] == provider) and (identity["id"] == username):
                    return
            raise RuntimeError(
                "An API key is set, and it is not associated with the username/provider "
                f"{username}/{provider}. Unset the API key first."
            )

        refresh_url = self.server_info["authentication"]["links"]["refresh_session"]
        csrf_token = self.http_client.cookies["tiled_csrf"]

        # If we are passed a username, we can check whether we already have
        # tokens stashed.
        if username is not None:
            token_directory = self._token_directory(provider, username)
            self.http_client.auth = TiledAuth(refresh_url, csrf_token, token_directory)
            # This will either:
            # * Use an access_token and succeed.
            # * Use a refresh_token to attempt refresh flow and succeed.
            # * Use a refresh_token to attempt refresh flow and fail, raise.
            # * Find no tokens and raise.
            try:
                self.whoami()
            except CannotRefreshAuthentication:
                # Continue below, where we will prompt for log in.
                self.http_client.auth = None
                if not prompt_for_reauthentication:
                    raise
            else:
                # We have a live session for the specified provider and username already.
                # No need to log in again.
                return

        if not prompt_for_reauthentication:
            raise CannotPrompt(
                """Authentication is needed but Tiled has detected that it is running
in a 'headless' context where it cannot prompt the user to provide
credentials in the stdin. Options:

- If Tiled has detected this wrongly, pass prompt_for_reauthentication=True
  to force it to prompt.
- Provide an API key in the environment variable TILED_API_KEY for Tiled to use.
"""
            )
        self.http_client.auth = None
        mode = spec["mode"]
        auth_endpoint = spec["links"]["auth_endpoint"]
        if mode == "password":
            if username:
                print(f"Username {username}")
            else:
                username = input("Username: ")
            password = getpass.getpass()
            form_data = {
                "grant_type": "password",
                "username": username,
                "password": password,
            }
            token_response = self.http_client.post(
                auth_endpoint, data=form_data, auth=None
            )
            handle_error(token_response)
            tokens = token_response.json()
        elif mode == "external":
            verification_response = self.http_client.post(
                auth_endpoint, json={}, auth=None
            )
            handle_error(verification_response)
            verification = verification_response.json()
            authorization_uri = verification["authorization_uri"]
            print(
                f"""
You have {int(verification['expires_in']) // 60} minutes visit this URL

  {authorization_uri}

and enter the code:

  {verification['user_code']}

"""
            )
            import webbrowser

            webbrowser.open(authorization_uri)
            deadline = verification["expires_in"] + time.monotonic()
            print("Waiting...", end="", flush=True)
            while True:
                time.sleep(verification["interval"])
                if time.monotonic() > deadline:
                    raise Exception("Deadline expired.")
                # Intentionally do not wrap this in handle_error(...).
                # Check status codes manually below.
                access_response = self.http_client.post(
                    verification["verification_uri"],
                    json={
                        "device_code": verification["device_code"],
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    },
                    auth=None,
                )
                if (access_response.status_code == 400) and (
                    access_response.json()["detail"]["error"] == "authorization_pending"
                ):
                    print(".", end="", flush=True)
                    continue
                handle_error(access_response)
                print("")
                break
            tokens = access_response.json()

        else:
            raise ValueError(f"Server has unknown authentication mechanism {mode!r}")
        username = tokens["identity"]["id"]
        token_directory = self._token_directory(provider, username)
        auth = TiledAuth(refresh_url, csrf_token, token_directory)
        auth.sync_set_token("access_token", tokens["access_token"])
        auth.sync_set_token("refresh_token", tokens["refresh_token"])
        self.http_client.auth = auth
        confirmation_message = spec.get("confirmation_message")
        if confirmation_message:
            print(confirmation_message.format(id=username))
        if set_default:
            set_default_identity(
                self.api_uri, username=username, provider=spec["provider"]
            )
        return spec, username

    def login(self, username=None, provider=None, prompt_for_reauthentication=UNSET):
        """
        Depending on the server's authentication method, this will prompt for username/password:

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party:

        >>> c.login()
        You have ... minutes visit this URL

        https://...

        and enter the code: XXXX-XXXX
        """
        self.authenticate(
            username, provider, prompt_for_reauthentication=prompt_for_reauthentication
        )
        # For programmatic access to the return values, use authenticate().
        # This returns None in order to provide a clean UX in an interpreter.
        return None

    def _token_directory(self, provider, username):
        # ~/.config/tiled/tokens/{host:port}/{provider}/{username}
        # with each templated element URL-encoded so it is a valid filename.
        return Path(
            self._token_cache,
            urllib.parse.quote_plus(str(self.api_uri)),
            urllib.parse.quote_plus(provider),
            urllib.parse.quote_plus(username),
        )

    def force_auth_refresh(self):
        """
        Execute refresh flow.

        This method is exposed for testing and debugging uses.

        It should never be necessary for the user to call. Refresh flow is
        automatically executed by tiled.client.auth.TiledAuth when the current
        access_token expires.
        """
        if self.http_client.auth is None:
            raise RuntimeError(
                "No authentication has been set up. Cannot reauthenticate."
            )
        refresh_token = self.http_client.auth.sync_get_token(
            "refresh_token", reload_from_disk=True
        )
        if refresh_token is None:
            raise CannotRefreshAuthentication("There is no refresh_token.")
        csrf_token = self.http_client.cookies["tiled_csrf"]
        refresh_request = build_refresh_request(
            self.http_client.auth.refresh_url,
            refresh_token,
            csrf_token,
        )
        token_response = self.http_client.send(refresh_request, auth=None)
        if token_response.status_code == 401:
            raise CannotRefreshAuthentication(
                "Session cannot be refreshed. Log in again."
            )
        handle_error(token_response)
        tokens = token_response.json()
        self.http_client.auth.sync_set_token("access_token", tokens["access_token"])
        self.http_client.auth.sync_set_token("refresh_token", tokens["refresh_token"])
        return tokens

    def whoami(self):
        "Return information about the currently-authenticated user or service."
        return handle_error(
            self.http_client.get(
                self.server_info["authentication"]["links"]["whoami"],
                headers={"Accept": MSGPACK_MIME_TYPE},
            )
        ).json()

    def logout(self):
        """
        Log out of the current session (if any).

        This method is idempotent.
        """
        if self.http_client.auth is None:
            return

        # Revoke the current session.
        refresh_token = self.http_client.auth.sync_get_token("refresh_token")
        handle_error(
            self.http_client.post(
                f"{self.api_uri}auth/session/revoke",
                json={"refresh_token": refresh_token},
                # Circumvent auth because this request is not authenticated.
                # The refresh_token in the body is the relevant proof, not the
                # 'Authentication' header.
                auth=None,
            )
        )

        # Clear on-disk state.
        self.http_client.auth.sync_clear_token("access_token")
        self.http_client.auth.sync_clear_token("refresh_token")

        # Clear in-memory state.
        self.http_client.headers.pop("Authorization", None)
        self.http_client.auth = None

    def revoke_session(self, session_id):
        """
        Revoke a Session so it cannot be refreshed.

        This may be done to ensure that a possibly-leaked refresh token cannot be used.
        """
        handle_error(
            self.http_client.delete(
                self.server_info["authentication"]["links"]["revoke_session"].format(
                    session_id=session_id
                ),
                headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
            )
        )


def _choose_identity_provider(providers, provider=None):
    if provider is not None:
        for spec in providers:
            if spec["provider"] == provider:
                break
        else:
            raise ValueError(
                f"No such provider {provider}. Choices are {[spec['provider'] for spec in providers]}"
            )
    else:
        if len(providers) == 1:
            # There is only one choice, so no need to prompt the user.
            (spec,) = providers
        else:
            while True:
                print("Authenticaiton providers:")
                for i, spec in enumerate(providers, start=1):
                    print(f"{i} - {spec['provider']}")
                raw_choice = input(
                    "Choose an authentication provider (or press Enter to escape): "
                )
                if not raw_choice:
                    print("No authentication provider chosen. Failed.")
                    break
                try:
                    choice = int(raw_choice)
                except TypeError:
                    print("Choice must be a number.")
                    continue
                try:
                    spec = providers[choice - 1]
                except IndexError:
                    print(f"Choice must be a number 1 through {len(providers)}.")
                    continue
                break
    return spec


class Admin:
    "Accessor for requests that require administrative privileges."

    def __init__(self, context):
        self.context = context
        self.base_url = context.server_info["links"]["self"]

    def list_principals(self, offset=0, limit=100):
        "List Principals (users and services) in the authenticaiton database."
        params = dict(offset=offset, limit=limit)
        return handle_error(
            self.context.http_client.get(
                f"{self.base_url}/auth/principal", params=params
            )
        ).json()

    def show_principal(self, uuid):
        "Show one Principal (user or service) in the authenticaiton database."
        return handle_error(
            self.context.http_client.get(f"{self.base_url}/auth/principal/{uuid}")
        ).json()

    def create_api_key(self, uuid, scopes=None, expires_in=None, note=None):
        """
        Generate a new API key for another user or service.

        Parameters
        ----------
        uuid : str
            Identify the principal -- the user or service
        scopes : Optional[List[str]]
            Restrict the access available to the API key by listing specific scopes.
            By default, this will have the same access as the principal.
        expires_in : Optional[int]
            Number of seconds until API key expires. If None,
            it will never expire or it will have the maximum lifetime
            allowed by the server.
        note : Optional[str]
            Description (for humans).
        """
        return handle_error(
            self.context.http_client.post(
                f"{self.base_url}/auth/principal/{uuid}/apikey",
                headers={"Accept": MSGPACK_MIME_TYPE},
                json={"scopes": scopes, "expires_in": expires_in, "note": note},
            )
        ).json()

    def create_service_principal(
        self,
        role,
    ):
        """
        Generate a new service principal.

        Parameters
        ----------
        role : str
            Specify the role (e.g. user or admin)
        """
        return handle_error(
            self.context.http_client.post(
                f"{self.base_url}/auth/principal",
                headers={"Accept": MSGPACK_MIME_TYPE},
                params={"role": role},
            )
        ).json()


class CannotPrompt(Exception):
    pass


def _can_prompt():
    "Infer whether the user can be prompted for a password or user code."

    if (not sys.__stdin__.closed) and sys.__stdin__.isatty():
        return True
    # In IPython (TerminalInteractiveShell) the above is true, but in
    # Jupyter (ZMQInteractiveShell) it is False.
    # Jupyter own mechanism for giving a prompt, so we always return
    # True if we detect IPython/Jupyter.
    if "IPython" in sys.modules:
        import IPython

        if IPython.get_ipython() is not None:
            return True
    return False


def _default_identity_filepath(api_uri):
    # Resolve this here, not at module scope, because the test suite
    # injects TILED_CACHE_DIR env var to use a temporary directory.
    TILED_CACHE_DIR = Path(
        os.getenv("TILED_CACHE_DIR", appdirs.user_cache_dir("tiled"))
    )
    return Path(
        TILED_CACHE_DIR, "default_identities", urllib.parse.quote_plus(str(api_uri))
    )


def set_default_identity(api_uri, provider, username):
    """
    Stash the identity used with this API so that we can reuse it by default.
    """
    filepath = _default_identity_filepath(api_uri)
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(filepath, "w") as file:
        json.dump({"username": username, "provider": provider}, file)


def get_default_identity(api_uri):
    """
    Look up the default identity to use with this API.
    """
    filepath = _default_identity_filepath(api_uri)
    if filepath.exists():
        return json.loads(filepath.read_text())


def clear_default_identity(api_uri):
    """
    Clear the cached default identity for this API.
    """
    filepath = _default_identity_filepath(api_uri)
    if filepath.exists():
        filepath.unlink()



import typer


def get_profile(name):
    from ..profiles import (
        get_default_profile_name,
        load_profiles,
        set_default_profile_name,
    )

    profiles = load_profiles()
    if name is None:
        # Use the default profile.
        # Raise if it is not set or if it is set but does not exit.
        name = get_default_profile_name()
        if name is None:
            typer.echo(
                """No default profile set. Use:

    tiled profile create ...

or

    tiled profile set-default ...

to set a create or choose a default profile or else specify a profile for this
particular command using

    tiled ... --profile=PROFILE
""",
                err=True,
            )
            raise typer.Abort()
        if name not in profiles:
            set_default_profile_name(None)
            typer.echo(
                f"""Default profile {name!r} does not exist. Clearing default. Use:

    tiled profile create ...

or

    tiled profile set-default ...

to set a create or choose a default profile or else specify a profile for this
particular command using

    tiled ... --profile=PROFILE
""",
                err=True,
            )
            raise typer.Abort()
    try:
        _, profile = profiles[name]
        if "direct" in profile:
            typer.echo(
                f"Profile {profile!r} uses in a direct (in-process) Tiled server "
                "and cannot be connected to from the CLI.",
                err=True,
            )
            typer.Abort()
    except KeyError:
        typer.echo(
            f"""Profile {name!r} could not be found. Use:

    tiled profile list

to list choices.""",
            err=True,
        )
        raise typer.Abort()
    return name, profile


def get_context(profile):
    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    context, _ = Context.from_any_uri(
        profile_content["uri"], verify=profile_content.get("verify", True)
    )
    if context.server_info["authentication"]["required"]:
        context.authenticate()
    return context



"""Changing top_level_metadata to btree_gin

Revision ID: 3db11ff95b6c
Revises: 0b033e7fbe30
Create Date: 2023-11-01 15:16:48.554420

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3db11ff95b6c"
down_revision = "0b033e7fbe30"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "postgresql":
        with op.get_context().autocommit_block():
            op.execute(sa.text("create extension IF NOT EXISTS btree_gin;"))
            op.drop_index("top_level_metadata", table_name="nodes")
            op.create_index(
                "top_level_metadata",
                "nodes",
                ["ancestors", "time_created", "id", "metadata"],
                postgresql_using="gin",
            )


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



import pytest

from ..adapters.mapping import MapAdapter
from ..client import Context
from ..server.app import build_app_from_config

# Basic authenticated server config
tree = MapAdapter({})
config = {
    "authentication": {
        "secret_keys": ["SECRET"],
        "providers": [
            {
                "provider": "toy",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {"users_to_passwords": {"alice": "secret1", "bob": "secret2"}},
            }
        ],
    },
    "trees": [
        {
            "tree": f"{__name__}:tree",
            "path": "/",
        },
    ],
}


@pytest.fixture
def context():
    with Context.from_app(build_app_from_config(config)) as context:
        yield context


def test_openapi_username_password_login(context):
    """
    Ensure that tokenUrl is a valid path.

    The tokenUrl is set manually because of when in the server lifecycle it is
    known, and the point of this test is to ensure it does not bit rot to some
    invalid path, because that has happened before.
    """
    response = context.http_client.get("/openapi.json")
    assert response.status_code == 200
    openapi = response.json()
    token_url = openapi["components"]["securitySchemes"]["OAuth2PasswordBearer"][
        "flows"
    ]["password"]["tokenUrl"]
    assert token_url in openapi["paths"]



"""
Adapted from https://raw.githubusercontent.com/obendidi/httpx-cache/main/httpx_cache/transport.py
in accordance with its BSD-3 license
"""
import typing as tp

import httpx

from .cache import Cache
from .cache_control import ByteStreamWrapper, CacheControl
from .logger import collect_request, collect_response, log_request, log_response, logger
from .utils import TiledResponse


class Transport(httpx.BaseTransport):
    """Custom transport, implementing caching and custom compression encodings.

    Args:
        transport (optional): an existing httpx transport, if no transport
            is given, defaults to an httpx.HTTPTransport with default args.
        cache (optional): cache to use with this transport, defaults to
            httpx_cache.DictCache
        cacheable_methods: methods that are allowed to be cached, defaults to ['GET']
        cacheable_status_codes: status codes that are allowed to be cached,
            defaults to: (200, 203, 300, 301, 308)
    """

    def __init__(
        self,
        *,
        transport: tp.Optional[httpx.BaseTransport] = None,
        cache: tp.Optional[Cache] = None,
        cacheable_methods: tp.Tuple[str, ...] = ("GET",),
        cacheable_status_codes: tp.Tuple[int, ...] = (200, 203, 300, 301, 308),
        always_cache: bool = False,
    ):
        self.controller = CacheControl(
            cacheable_methods=cacheable_methods,
            cacheable_status_codes=cacheable_status_codes,
            always_cache=always_cache,
        )
        self.transport = transport or httpx.HTTPTransport()
        self.cache = cache

    def close(self) -> None:
        self.transport.close()
        if self.cache is not None:
            self.cache.close()

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        # check if request is cacheable
        if (self.cache is not None) and self.controller.is_request_cacheable(request):
            if __debug__:
                logger.debug("Checking cache for: %s", request)
            cached_response = self.cache.get(request)
            if cached_response is not None:
                if self.controller.is_response_fresh(
                    request=request, response=cached_response
                ):
                    if not self.controller.needs_revalidation(
                        request=request, response=cached_response
                    ):
                        if __debug__:
                            logger.debug("Using cached response for: %s", request)
                            log_request(request)
                            collect_request(request)
                        return cached_response
                    if __debug__:
                        logger.debug("Revalidating cached response for: %s", request)
                    request.headers["If-None-Match"] = cached_response.headers["ETag"]
                else:
                    if __debug__:
                        logger.debug("Cached response is stale, deleting: %s", request)
                    self.cache.delete(request)
            else:
                if __debug__:
                    logger.debug(
                        "No valid cached response found in cache for: %s", request
                    )

        # Call original transport
        if __debug__:
            log_request(request)
            collect_request(request)
        response = self.transport.handle_request(request)
        response.__class__ = TiledResponse
        response.request = request
        if __debug__:
            # Log the actual server traffic, not the cached response.
            log_response(response)
            # But, below _collect_ the response with the content in it.

        if self.cache is not None:
            if response.status_code == 304:
                if __debug__:
                    logger.debug(
                        "Server validated as fresh cached entry for: %s", request
                    )
                    collect_response(cached_response)
                return cached_response

            if self.controller.is_response_cacheable(
                request=request, response=response
            ):
                if self.cache.readonly:
                    if __debug__:
                        logger.debug("Cache is read-only; will not store")
                elif not self.cache.write_safe():
                    if __debug__:
                        logger.debug(
                            "Cannot write to cache from another thread; will not store"
                        )
                else:
                    if hasattr(response, "_content"):
                        is_stored = self.cache.set(request=request, response=response)
                        if __debug__:
                            if is_stored:
                                logger.debug("Caching response for: %s", request)
                            else:
                                logger.debug(
                                    "Declined to store large response for: %s", request
                                )
                    else:
                        # Wrap the response with cache callback:
                        def _callback(content: bytes) -> None:
                            is_stored = self.cache.set(
                                request=request, response=response, content=content
                            )
                            if __debug__:
                                if is_stored:
                                    logger.debug("Caching response for: %s", request)
                                else:
                                    logger.debug(
                                        "Declined to store large response for: %s",
                                        request,
                                    )

                        response.stream = ByteStreamWrapper(
                            stream=response.stream, callback=_callback  # type: ignore
                        )
        if __debug__:
            collect_response(response)
        return response


# For when we implement an Async client
#
# class AsyncCacheControlTransport(httpx.AsyncBaseTransport):
#     """Async CacheControl transport for httpx_cache.
#
#     Args:
#         transport (optional): an existing httpx async-transport, if no transport
#             is given, defaults to an httpx.AsyncHTTPTransport with default args.
#         cache (optional): cache to use with this transport, defaults to
#             httpx_cache.DictCache
#         cacheable_methods: methods that are allowed to be cached, defaults to ['GET']
#         cacheable_status_codes: status codes that are allowed to be cached,
#             defaults to: (200, 203, 300, 301, 308)
#     """
#
#     def __init__(
#         self,
#         *,
#         transport: tp.Optional[httpx.AsyncBaseTransport] = None,
#         cache: tp.Optional[BaseCache] = None,
#         cacheable_methods: tp.Tuple[str, ...] = ("GET",),
#         cacheable_status_codes: tp.Tuple[int, ...] = (200, 203, 300, 301, 308),
#         always_cache: bool = False,
#     ):
#         self.controller = CacheControl(
#             cacheable_methods=cacheable_methods,
#             cacheable_status_codes=cacheable_status_codes,
#             always_cache=always_cache,
#         )
#         self.transport = transport or httpx.AsyncHTTPTransport()
#         self.cache = cache or DictCache()
#
#     async def aclose(self) -> None:
#         await self.cache.aclose()
#         await self.transport.aclose()
#
#     async def handle_async_request(self, request: httpx.Request) -> TiledResponse:
#         # check if request is cacheable
#         if self.controller.is_request_cacheable(request):
#             logger.debug(f"Checking cache for: %s", request)
#             cached_response = await self.cache.aget(request)
#             if cached_response is not None:
#                 logger.debug(f"Found cached response for: %s", request)
#                 if self.controller.is_response_fresh(
#                     request=request, response=cached_response
#                 ):
#                     setattr(cached_response, "from_cache", True)
#                     return cached_response
#                 else:
#                     logger.debug(f"Cached response is stale, deleting: %s", request)
#                     await self.cache.adelete(request)
#
#         # Request is not in cache, call original transport
#         response = await self.transport.handle_async_request(request)
#
#         if self.controller.is_response_cacheable(request=request, response=response):
#             if hasattr(response, "_content"):
#                 logger.debug(f"Caching response for: %s", request)
#                 await self.cache.aset(request=request, response=response)
#             else:
#                 # Wrap the response with cache callback:
#                 async def _callback(content: bytes) -> None:
#                     logger.debug(f"Caching response for: %s", request)
#                     await self.cache.aset(
#                         request=request, response=response, content=content
#                     )
#
#                 response.stream = ByteStreamWrapper(
#                     stream=response.stream, callback=_callback  # type: ignore
#                 )
#         setattr(response, "from_cache", False)
#         return response



from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine

from ..server.settings import get_settings

# A given process probably only has one of these at a time, but we
# key on database_settings just case in some testing context or something
# we have two servers running in the same process.
_connection_pools = {}


def open_database_connection_pool(database_settings):
    connect_args = {}
    kwargs = {}  # extra kwargs passed to create_engine
    # kwargs["pool_size"] = database_settings.pool_size
    # kwargs["pool_pre_ping"] = database_settings.pool_pre_ping
    # kwargs["max_overflow"] = database_settings.max_overflow
    engine = create_async_engine(
        database_settings.uri, connect_args=connect_args, **kwargs
    )
    _connection_pools[database_settings] = engine
    return engine


async def close_database_connection_pool(database_settings):
    engine = _connection_pools.pop(database_settings, None)
    if engine is not None:
        await engine.dispose()


async def get_database_engine(settings=Depends(get_settings)):
    # Special case for single-user mode
    if settings.database_uri is None:
        return None
    try:
        return _connection_pools[settings.database_settings]
    except KeyError:
        raise RuntimeError(
            f"Could not find connection pool for {settings.database_settings}"
        )


async def get_database_session(engine=Depends(get_database_engine)):
    # Special case for single-user mode
    if engine is None:
        yield None
    else:
        async with AsyncSession(
            engine, autoflush=False, expire_on_commit=False
        ) as session:
            yield session



import dataclasses
import platform
import random
from pathlib import Path

import h5py
import numpy
import pytest
import tifffile
import yaml

from ..adapters.hdf5 import HDF5Adapter
from ..adapters.tiff import TiffAdapter
from ..catalog import in_memory
from ..catalog.register import (
    Settings,
    create_node_safe,
    group_tiff_sequences,
    identity,
    register,
    register_tiff_sequence,
    skip_all,
    strip_suffixes,
)
from ..catalog.utils import ensure_uri
from ..client import Context, from_context
from ..examples.generate_files import data, df1, generate_files
from ..server.app import build_app
from ..server.schemas import Asset, DataSource, Management
from ..utils import path_from_uri


@pytest.fixture
def example_data_dir(tmpdir_factory):
    """
    Generate a temporary directory with example files.

    The tmpdir_factory fixture ensures that this directory is cleaned up at test exit.
    """
    tmpdir = tmpdir_factory.mktemp("example_files")
    generate_files(tmpdir)
    return tmpdir


@pytest.mark.xfail(
    platform.system() == "Windows",
    reason="file cannot be removed while being used",
    raises=PermissionError,
)
@pytest.mark.asyncio
async def test_collision(example_data_dir, tmpdir):
    """Test that files which produce key collisions are ignored until the collision is resolved."""
    # Add a.tiff which will collide with a.tif.
    p = Path(example_data_dir, "a.tiff")
    tifffile.imwrite(str(p), data)

    catalog = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(catalog)) as context:
        await register(catalog, example_data_dir)

        client = from_context(context)

        # And omits the colliding entries.
        assert "a" not in client

        # Resolve the collision.
        p.unlink()

        # Re-run registration; entry should be there now.
        await register(catalog, example_data_dir)
        assert "a" in client


@pytest.mark.parametrize(
    ("filename", "expected"), [("a.txt", "a"), ("a.tar.gz", "a"), ("a", "a")]
)
def test_strip_suffixes(filename, expected):
    actual = strip_suffixes(filename)
    assert actual == expected


@pytest.mark.asyncio
async def test_same_filename_separate_directory(tmpdir):
    "Two files with the same name in separate directories should not collide."
    Path(tmpdir, "one").mkdir()
    Path(tmpdir, "two").mkdir()
    df1.to_csv(Path(tmpdir, "one", "a.csv"))
    df1.to_csv(Path(tmpdir, "two", "a.csv"))
    catalog = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(catalog)) as context:
        await register(catalog, tmpdir)
        client = from_context(context)
        assert "a" in client["one"]
        assert "a" in client["two"]


@pytest.mark.asyncio
async def test_mimetype_detection_hook(tmpdir):
    content = "a, b, c\n1, 2 ,3\n4, 5, 6\n"
    with open(Path(tmpdir / "a0"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "b0"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "c.csv"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "a.0.asfwoeijviojefeiofw"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "b.0.asfwoeijviojefeiofw"), "w") as file:
        file.write(content)

    def detect_mimetype(path, mimetype):
        filename = Path(path).name
        # If detection based on file extension worked,
        # we should get that in the mimetype. Otherwise,
        # mimetype should be None.
        if filename.endswith(".csv"):
            assert mimetype == "text/csv"
        else:
            assert mimetype is None
        if filename.startswith("a"):
            return "text/csv"
        return mimetype

    catalog = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(catalog)) as context:
        await register(
            catalog,
            tmpdir,
            mimetype_detection_hook=detect_mimetype,
            key_from_filename=identity,
        )
        client = from_context(context)
        assert set(client) == {"a0", "a.0.asfwoeijviojefeiofw", "c.csv"}


@pytest.mark.asyncio
async def test_skip_all_in_combination(tmpdir):
    "Using skip_all should override defaults, but not hinder other walkers"
    df1.to_csv(Path(tmpdir, "a.csv"))
    Path(tmpdir, "one").mkdir()
    df1.to_csv(Path(tmpdir, "one", "a.csv"))

    for i in range(2):
        tifffile.imwrite(Path(tmpdir, "one", f"image{i:05}.tif"), data)

    catalog = in_memory(writable_storage=tmpdir)
    # By default, both file and tiff sequence are registered.
    with Context.from_app(build_app(catalog)) as context:
        await register(catalog, tmpdir)
        client = from_context(context)
        assert "a" in client
        assert "a" in client["one"]
        assert "image" in client["one"]

    # With skip_all, directories and tiff sequence are registered, but individual files are not
    with Context.from_app(build_app(catalog)) as context:
        await register(catalog, tmpdir, walkers=[group_tiff_sequences, skip_all])
        client = from_context(context)
        assert list(client) == ["one"]
        assert "image" in client["one"]


@pytest.mark.asyncio
async def test_tiff_seq_custom_sorting(tmpdir):
    "Register TIFFs that are not in alphanumeric order."
    N = 10
    ordering = list(range(N))
    random.Random(0).shuffle(ordering)
    files = []
    for i in ordering:
        file = Path(tmpdir, f"image{i:05}.tif")
        files.append(file)
        # data is a block of ones
        tifffile.imwrite(file, i * data)

    settings = Settings.init()
    catalog = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(catalog)) as context:
        await register_tiff_sequence(
            catalog,
            "image",
            files,
            settings,
        )
        client = from_context(context)
        # We are being a bit clever here.
        # Each image in this image series has pixels with a constant value, and
        # that value matches the image's position in the sequence enumerated by
        # `ordering`. We pick out one pixel and check that its value matches
        # the corresponding value in `ordering`.
        actual = list(client["image"][:, 0, 0])
        assert actual == ordering


@pytest.mark.asyncio
async def test_image_file_with_sidecar_metadata_file(tmpdir):
    "Create one Node from two different types of files."
    MIMETYPE = "multipart/related;type=application/x-tiff-with-yaml"
    image_filepath = Path(tmpdir, "image.tif")
    tifffile.imwrite(image_filepath, data)
    metadata_filepath = Path(tmpdir, "metadata.yml")
    metadata = {"test_key": 3.0}
    with open(metadata_filepath, "w") as file:
        yaml.dump(metadata, file)

    def read_tiff_with_yaml_metadata(image_uri, metadata_uri, metadata=None, **kwargs):
        with open(path_from_uri(metadata_uri)) as file:
            metadata = yaml.safe_load(file)
        return TiffAdapter(image_uri, metadata=metadata, **kwargs)

    catalog = in_memory(
        writable_storage=tmpdir,
        adapters_by_mimetype={MIMETYPE: read_tiff_with_yaml_metadata},
    )
    with Context.from_app(build_app(catalog)) as context:
        adapter = read_tiff_with_yaml_metadata(
            ensure_uri(image_filepath), ensure_uri(metadata_filepath)
        )
        await create_node_safe(
            catalog,
            key="image",
            structure_family=adapter.structure_family,
            metadata=dict(adapter.metadata()),
            specs=adapter.specs,
            data_sources=[
                DataSource(
                    mimetype=MIMETYPE,
                    structure=dataclasses.asdict(adapter.structure()),
                    parameters={},
                    management=Management.external,
                    assets=[
                        Asset(
                            data_uri=ensure_uri(metadata_filepath),
                            is_directory=False,
                            parameter="metadata_uri",
                        ),
                        Asset(
                            data_uri=ensure_uri(image_filepath),
                            is_directory=False,
                            parameter="image_uri",
                        ),
                    ],
                )
            ],
        )
        client = from_context(context)
        assert numpy.array_equal(data, client["image"][:])
        assert client["image"].metadata["test_key"] == 3.0


@pytest.mark.asyncio
async def test_hdf5_virtual_datasets(tmpdir):
    # A virtual database comprises one master file and N data files. The master
    # file must be handed to the Adapter for opening. The data files are not
    # handled directly by the Adapter but they still ought to be tracked as
    # Assets for purposes of data movement, accounting for data size, etc.
    # This is why they are Assets with parameter=NULL/None, Assets not used
    # directly by the Adapter.

    # One could do one-dataset-per-directory. But like TIFF series in practice
    # they are often mixed, so we address that general case and track them at
    # the per-file level.

    # Contrast this to Zarr, where the files involves are always bundled by
    # directory. We track Zarr at the directory level.

    layout = h5py.VirtualLayout(shape=(4, 100), dtype="i4")

    data_filepaths = []
    for n in range(1, 5):
        filepath = Path(tmpdir, f"{n}.h5")
        data_filepaths.append(filepath)
        vsource = h5py.VirtualSource(filepath, "data", shape=(100,))
        layout[n - 1] = vsource

    # Add virtual dataset to output file
    filepath = Path(tmpdir, "VDS.h5")
    with h5py.File(filepath, "w", libver="latest") as file:
        file.create_virtual_dataset("data", layout, fillvalue=-5)

    assets = [
        Asset(
            data_uri=str(ensure_uri(str(fp))),
            is_directory=False,
            parameter=None,  # an indirect dependency
        )
        for fp in data_filepaths
    ]
    assets.append(
        Asset(
            data_uri=ensure_uri(filepath),
            is_directory=False,
            parameter="data_uri",
        )
    )
    catalog = in_memory(writable_storage=tmpdir)
    with Context.from_app(build_app(catalog)) as context:
        adapter = HDF5Adapter.from_uri(ensure_uri(filepath))
        await create_node_safe(
            catalog,
            key="VDS",
            structure_family=adapter.structure_family,
            metadata=dict(adapter.metadata()),
            specs=adapter.specs,
            data_sources=[
                DataSource(
                    mimetype="application/x-hdf5",
                    structure=None,
                    parameters={},
                    management=Management.external,
                    assets=assets,
                )
            ],
        )
        client = from_context(context)
        client["VDS"]["data"][:]



import enum
from dataclasses import dataclass
from typing import Optional, Tuple, Union


class SparseLayout(str, enum.Enum):
    # Only COO is currently supported, but this lays a path
    # for adding other layouts like CSC, CSR, etc. in the future.
    COO = "COO"


@dataclass
class COOStructure:
    chunks: Tuple[Tuple[int, ...], ...]  # tuple-of-tuples-of-ints like ((3,), (3,))
    shape: Tuple[int, ...]  # tuple of ints like (3, 3)
    dims: Optional[Tuple[str, ...]] = None  # None or tuple of names like ("x", "y")
    resizable: Union[bool, Tuple[bool, ...]] = False
    layout: SparseLayout = SparseLayout.COO
    # TODO Include fill_value?

    @classmethod
    def from_json(cls, structure):
        return cls(
            chunks=tuple(map(tuple, structure["chunks"])),
            shape=tuple(structure["shape"]),
            dims=structure["dims"],
            resizable=structure.get("resizable", False),
        )


# This may be extended to a Union of structures if more are added.
SparseStructure = COOStructure



import uuid

import numpy as np
import pandas as pd
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..queries import Key
from ..server.app import build_app

values = ["a", "b", "c"]
counts = [10, 5, 7]

mapping = {}

for value, count in zip(values, counts):
    for _ in range(count):
        mapping[str(uuid.uuid4())] = MapAdapter({}, metadata={"foo": {"bar": value}})

# items which do not contain the queries metadata should not effect the results
for _ in range(10):
    mapping[str(uuid.uuid4())] = ArrayAdapter.from_array(
        np.ones(10), metadata={}, specs=["MyArray"]
    )

# Added additional field in metadata to implement consecutive search and distinct queries
for i in range(10):
    if i < 5:
        group = "A"
    else:
        group = "B"

    if i % 2 == 0:
        subgroup = "even"
        specs = ["MyDataFrame", "ExtendedSpec"]
    else:
        subgroup = "odd"
        specs = ["MyDataFrame"]

    if i == 0:
        tag = "Zero"
    else:
        for j in range(2, int(i / 2) + 1):
            if (i % j) == 0:
                tag = "NotPrime"
                break
        else:
            tag = "Prime"

    mapping[str(uuid.uuid4())] = DataFrameAdapter.from_pandas(
        pd.DataFrame({"a": np.ones(10)}),
        metadata={"group": group, "subgroup": subgroup, "tag": tag},
        specs=specs,
        npartitions=1,
    )

tree = MapAdapter(mapping)


@pytest.fixture(scope="module")
def context():
    app = build_app(tree)
    with Context.from_app(app) as context:
        yield context


def test_distinct(context):
    client = from_context(context)
    # test without counts
    distinct = client.distinct(
        "foo.bar", structure_families=True, specs=True, counts=False
    )
    expected = {
        "metadata": {"foo.bar": [{"value": v, "count": None} for v in values]},
        "specs": [
            {"value": [], "count": None},
            {"value": ["MyArray"], "count": None},
            {"value": ["MyDataFrame", "ExtendedSpec"], "count": None},
            {"value": ["MyDataFrame"], "count": None},
        ],
        "structure_families": [
            {"value": "container", "count": None},
            {"value": "array", "count": None},
            {"value": "table", "count": None},
        ],
    }

    assert distinct["metadata"] == expected["metadata"]
    assert distinct["specs"] == expected["specs"]
    assert distinct["structure_families"] == expected["structure_families"]

    # test with counts
    distinct = client.distinct(
        "foo.bar", structure_families=True, specs=True, counts=True
    )
    expected = {
        "metadata": {
            "foo.bar": [{"value": v, "count": c} for v, c in zip(values, counts)]
        },
        "specs": [
            {"value": [], "count": 22},
            {"value": ["MyArray"], "count": 10},
            {"value": ["MyDataFrame", "ExtendedSpec"], "count": 5},
            {"value": ["MyDataFrame"], "count": 5},
        ],
        "structure_families": [
            {"value": "container", "count": 22},
            {"value": "array", "count": 10},
            {"value": "table", "count": 10},
        ],
    }

    assert distinct["metadata"] == expected["metadata"]
    assert distinct["specs"] == expected["specs"]
    assert distinct["structure_families"] == expected["structure_families"]

    # test with no matches
    distinct = client.distinct("baz", counts=True)
    expected = {"baz": []}
    assert distinct["metadata"] == expected


def test_search_distinct(context):
    client = from_context(context)
    distinct = (
        client.search(Key("group") == "A")
        .search(Key("subgroup") == "odd")
        .distinct("tag", counts=True)
    )

    expected = {
        "metadata": {
            "tag": [
                {"value": "Prime", "count": 2},
            ],
        },
    }

    assert distinct["metadata"] == expected["metadata"]



import dataclasses
import inspect
from datetime import datetime, timedelta
from functools import partial
from typing import Any, List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from jmespath.exceptions import JMESPathError
from pydantic import BaseSettings

from .. import __version__
from ..structures.core import StructureFamily
from ..utils import ensure_awaitable
from ..validation_registration import ValidationError
from . import schemas
from .authentication import Mode, get_authenticators, get_current_principal
from .core import (
    DEFAULT_PAGE_SIZE,
    DEPTH_LIMIT,
    MAX_PAGE_SIZE,
    NoEntry,
    UnsupportedMediaTypes,
    WrongTypeForRoute,
    apply_search,
    construct_data_response,
    construct_entries_response,
    construct_resource,
    construct_revisions_response,
    json_or_msgpack,
    resolve_media_type,
)
from .dependencies import (
    SecureEntry,
    block,
    expected_shape,
    get_deserialization_registry,
    get_query_registry,
    get_serialization_registry,
    get_validation_registry,
    slice_,
)
from .settings import get_settings
from .utils import filter_for_access, get_base_url, record_timing

router = APIRouter()


@router.get("/", response_model=schemas.About)
async def about(
    request: Request,
    settings: BaseSettings = Depends(get_settings),
    authenticators=Depends(get_authenticators),
    serialization_registry=Depends(get_serialization_registry),
    query_registry=Depends(get_query_registry),
    # This dependency is here because it runs the code that moves
    # API key from the query parameter to a cookie (if it is valid).
    principal=Security(get_current_principal, scopes=[]),
):
    # TODO The lazy import of entry modules and serializers means that the
    # lists of formats are not populated until they are first used. Not very
    # helpful for discovery! The registration can be made non-lazy, while the
    # imports of the underlying I/O libraries themselves (openpyxl, pillow,
    # etc.) can remain lazy.
    request.state.endpoint = "about"
    base_url = get_base_url(request)
    authentication = {
        "required": not settings.allow_anonymous_access,
    }
    provider_specs = []
    for provider, authenticator in authenticators.items():
        if authenticator.mode == Mode.password:
            spec = {
                "provider": provider,
                "mode": authenticator.mode.value,
                "links": {
                    "auth_endpoint": f"{base_url}/auth/provider/{provider}/token"
                },
                "confirmation_message": getattr(
                    authenticator, "confirmation_message", None
                ),
            }
        elif authenticator.mode == Mode.external:
            spec = {
                "provider": provider,
                "mode": authenticator.mode.value,
                "links": {
                    "auth_endpoint": f"{base_url}/auth/provider/{provider}/authorize"
                },
                "confirmation_message": getattr(
                    authenticator, "confirmation_message", None
                ),
            }
        else:
            # It should be impossible to reach here.
            assert False
        provider_specs.append(spec)
    if provider_specs:
        # If there are *any* authenticaiton providers, these
        # endpoints will be added.
        authentication["links"] = {
            "whoami": f"{base_url}/auth/whoami",
            "apikey": f"{base_url}/auth/apikey",
            "refresh_session": f"{base_url}/auth/session/refresh",
            "revoke_session": f"{base_url}/auth/session/revoke/{{session_id}}",
            "logout": f"{base_url}/auth/logout",
        }
    authentication["providers"] = provider_specs

    return json_or_msgpack(
        request,
        schemas.About(
            library_version=__version__,
            api_version=0,
            formats={
                structure_family: list(
                    serialization_registry.media_types(structure_family)
                )
                for structure_family in serialization_registry.structure_families
            },
            aliases={
                structure_family: serialization_registry.aliases(structure_family)
                for structure_family in serialization_registry.structure_families
            },
            queries=list(query_registry.name_to_query_type),
            authentication=authentication,
            links={
                "self": base_url,
                "documentation": f"{base_url}/docs",
            },
            meta={"root_path": request.scope.get("root_path") or "" + "/api"},
        ).dict(),
        expires=datetime.utcnow() + timedelta(seconds=600),
    )


async def search(
    request: Request,
    path: str,
    fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
    select_metadata: Optional[str] = Query(None),
    offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
    limit: Optional[int] = Query(
        DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
    ),
    sort: Optional[str] = Query(None),
    max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
    omit_links: bool = Query(False),
    show_sources: bool = Query(False),
    entry: Any = SecureEntry(scopes=["read:metadata"]),
    query_registry=Depends(get_query_registry),
    principal: str = Depends(get_current_principal),
    **filters,
):
    request.state.endpoint = "search"
    if entry.structure_family != StructureFamily.container:
        raise WrongTypeForRoute("This is not a Node; it cannot be searched or listed.")
    entry = filter_for_access(
        entry, principal, ["read:metadata"], request.state.metrics
    )
    try:
        resource, metadata_stale_at, must_revalidate = await construct_entries_response(
            query_registry,
            entry,
            "/search",
            path,
            offset,
            limit,
            fields,
            select_metadata,
            omit_links,
            show_sources,
            filters,
            sort,
            get_base_url(request),
            resolve_media_type(request),
            max_depth=max_depth,
        )
        # We only get one Expires header, so if different parts
        # of this response become stale at different times, we
        # cite the earliest one.
        entries_stale_at = getattr(entry, "entries_stale_at", None)
        headers = {}
        if (metadata_stale_at is None) or (entries_stale_at is None):
            expires = None
        else:
            expires = min(metadata_stale_at, entries_stale_at)
        if must_revalidate:
            headers["Cache-Control"] = "must-revalidate"
        return json_or_msgpack(
            request,
            resource.dict(),
            expires=expires,
            headers=headers,
        )
    except NoEntry:
        raise HTTPException(status_code=404, detail="No such entry.")
    except WrongTypeForRoute as err:
        raise HTTPException(status_code=404, detail=err.args[0])
    except JMESPathError as err:
        raise HTTPException(
            status_code=400,
            detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
        )


async def distinct(
    request: Request,
    structure_families: bool = False,
    specs: bool = False,
    metadata: Optional[List[str]] = Query(default=[]),
    counts: bool = False,
    entry: Any = SecureEntry(scopes=["read:metadata"]),
    query_registry=Depends(get_query_registry),
    **filters,
):
    if hasattr(entry, "get_distinct"):
        filtered = await apply_search(entry, filters, query_registry)
        distinct = await ensure_awaitable(
            filtered.get_distinct, metadata, structure_families, specs, counts
        )

        return json_or_msgpack(
            request, schemas.GetDistinctResponse.parse_obj(distinct).dict()
        )
    else:
        raise HTTPException(
            status_code=405, detail="This node does not support distinct."
        )


def patch_route_signature(route, query_registry):
    """
    This is done dynamically at router startup.

    We check the registry of known search query types, which is user
    configurable, and use that to define the allowed HTTP query parameters for
    this route.

    Take a route that accept unspecified search queries as **filters.
    Return a wrapped version of the route that has the supported
    search queries explicitly spelled out in the function signature.

    This has no change in the actual behavior of the function,
    but it enables FastAPI to generate good OpenAPI documentation
    showing the supported search queries.

    """

    # Build a wrapper so that we can modify the signature
    # without mutating the wrapped original.

    async def route_with_sig(*args, **kwargs):
        return await route(*args, **kwargs)

    # Black magic here! FastAPI bases its validation and auto-generated swagger
    # documentation on the signature of the route function. We do not know what
    # that signature should be at compile-time. We only know it once we have a
    # chance to check the user-configurable registry of query types. Therefore,
    # we modify the signature here, at runtime, just before handing it to
    # FastAPI in the usual way.

    # When FastAPI calls the function with these added parameters, they will be
    # accepted via **filters.

    # Make a copy of the original parameters.
    signature = inspect.signature(route)
    parameters = list(signature.parameters.values())
    # Drop the **filters parameter from the signature.
    del parameters[-1]
    # Add a parameter for each field in each type of query.
    for name, query in query_registry.name_to_query_type.items():
        for field in dataclasses.fields(query):
            # The structured "alias" here is based on
            # https://mglaman.dev/blog/using-json-router-query-your-search-router-indexes
            if getattr(field.type, "__origin__", None) is list:
                field_type = str
            else:
                field_type = field.type
            injected_parameter = inspect.Parameter(
                name=f"filter___{name}___{field.name}",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=Query(None, alias=f"filter[{name}][condition][{field.name}]"),
                annotation=Optional[List[field_type]],
            )
            parameters.append(injected_parameter)
    route_with_sig.__signature__ = signature.replace(parameters=parameters)
    # End black magic

    return route_with_sig


@router.get(
    "/metadata/{path:path}",
    response_model=schemas.Response[
        schemas.Resource[schemas.NodeAttributes, dict, dict], dict, dict
    ],
)
async def metadata(
    request: Request,
    path: str,
    fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
    select_metadata: Optional[str] = Query(None),
    max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
    omit_links: bool = Query(False),
    show_sources: bool = Query(False),
    entry: Any = SecureEntry(scopes=["read:metadata"]),
    root_path: bool = Query(False),
):
    "Fetch the metadata and structure information for one entry."

    request.state.endpoint = "metadata"
    base_url = get_base_url(request)
    path_parts = [segment for segment in path.split("/") if segment]
    try:
        resource = await construct_resource(
            base_url,
            path_parts,
            entry,
            fields,
            select_metadata,
            omit_links,
            show_sources,
            resolve_media_type(request),
            max_depth=max_depth,
        )
    except JMESPathError as err:
        raise HTTPException(
            status_code=400,
            detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
        )
    meta = {"root_path": request.scope.get("root_path") or "/"} if root_path else {}
    return json_or_msgpack(
        request,
        schemas.Response(data=resource, meta=meta).dict(),
        expires=getattr(entry, "metadata_stale_at", None),
    )


@router.get(
    "/array/block/{path:path}", response_model=schemas.Response, name="array block"
)
async def array_block(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    block=Depends(block),
    slice=Depends(slice_),
    expected_shape=Depends(expected_shape),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a chunk of array-like data.
    """
    if entry.structure_family == "array":
        shape = entry.structure().shape
    elif entry.structure_family == "sparse":
        shape = entry.structure().shape
    else:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /array/block route.",
        )
    # Check that block dimensionality matches array dimensionality.
    ndim = len(shape)
    if len(block) != ndim:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Block parameter must have {ndim} comma-separated parameters, "
                f"corresponding to the dimensions of this {ndim}-dimensional array."
            ),
        )
    if block == ():
        # Handle special case of numpy scalar.
        if shape != ():
            raise HTTPException(
                status_code=400,
                detail=f"Requested scalar but shape is {entry.structure().shape}",
            )
        with record_timing(request.state.metrics, "read"):
            array = await ensure_awaitable(entry.read)
    else:
        try:
            with record_timing(request.state.metrics, "read"):
                array = await ensure_awaitable(entry.read_block, block, slice)
        except IndexError:
            raise HTTPException(status_code=400, detail="Block index out of range")
        if (expected_shape is not None) and (expected_shape != array.shape):
            raise HTTPException(
                status_code=400,
                detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
            )
    if array.nbytes > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                entry.structure_family,
                serialization_registry,
                array,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        # raise HTTPException(status_code=406, detail=", ".join(err.supported))
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/array/full/{path:path}", response_model=schemas.Response, name="full array"
)
async def array_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    slice=Depends(slice_),
    expected_shape=Depends(expected_shape),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a slice of array-like data.
    """
    structure_family = entry.structure_family
    if structure_family not in {"array", "sparse"}:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /array/full route.",
        )
    # Deferred import because this is not a required dependency of the server
    # for some use cases.
    import numpy

    try:
        with record_timing(request.state.metrics, "read"):
            array = await ensure_awaitable(entry.read, slice)
        if structure_family == "array":
            array = numpy.asarray(array)  # Force dask or PIMS or ... to do I/O.
    except IndexError:
        raise HTTPException(status_code=400, detail="Block index out of range")
    if (expected_shape is not None) and (expected_shape != array.shape):
        raise HTTPException(
            status_code=400,
            detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
        )
    if array.nbytes > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                structure_family,
                serialization_registry,
                array,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/table/partition/{path:path}",
    response_model=schemas.Response,
    name="table partition",
)
async def table_partition(
    request: Request,
    partition: int,
    entry=SecureEntry(scopes=["read:data"]),
    field: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a partition (continuous block of rows) from a DataFrame.
    """
    if entry.structure_family != StructureFamily.table:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /table/partition route.",
        )
    try:
        # The singular/plural mismatch here of "fields" and "field" is
        # due to the ?field=A&field=B&field=C... encodes in a URL.
        with record_timing(request.state.metrics, "read"):
            df = await ensure_awaitable(entry.read_partition, partition, field)
    except IndexError:
        raise HTTPException(status_code=400, detail="Partition out of range")
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    if df.memory_usage().sum() > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Select a subset of the columns ('?field=...') to "
                "request a smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                StructureFamily.table,
                serialization_registry,
                df,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/table/full/{path:path}",
    response_model=schemas.Response,
    name="full 'table' data",
)
async def table_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    column: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch the data for the given table.
    """
    if entry.structure_family != StructureFamily.table:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /table/full route.",
        )
    try:
        with record_timing(request.state.metrics, "read"):
            data = await ensure_awaitable(entry.read, column)
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    if data.memory_usage().sum() > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Select a subset of the columns to "
                "request a smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                entry.structure_family,
                serialization_registry,
                data,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
                filter_for_access=None,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/container/full/{path:path}",
    response_model=schemas.Response,
    name="full 'container' metadata and data",
)
async def container_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    principal: str = Depends(get_current_principal),
    field: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
):
    """
    Fetch the data for the given container.
    """
    if entry.structure_family != StructureFamily.container:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /container/full route.",
        )
    try:
        with record_timing(request.state.metrics, "read"):
            data = await ensure_awaitable(entry.read, field)
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    curried_filter = partial(
        filter_for_access,
        principal=principal,
        scopes=["read:data"],
        metrics=request.state.metrics,
    )
    # TODO Walk node to determine size before handing off to serializer.
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                entry.structure_family,
                serialization_registry,
                data,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
                filter_for_access=curried_filter,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/node/full/{path:path}",
    response_model=schemas.Response,
    name="full 'container' or 'table'",
    deprecated=True,
)
async def node_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    principal: str = Depends(get_current_principal),
    field: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch the data below the given node.
    """
    try:
        with record_timing(request.state.metrics, "read"):
            data = await ensure_awaitable(entry.read, field)
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    if (entry.structure_family == StructureFamily.table) and (
        data.memory_usage().sum() > settings.response_bytesize_limit
    ):
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Select a subset of the columns ('?field=...') to "
                "request a smaller chunks."
            ),
        )
    if entry.structure_family == StructureFamily.container:
        curried_filter = partial(
            filter_for_access,
            principal=principal,
            scopes=["read:data"],
            metrics=request.state.metrics,
        )
    else:
        curried_filter = None
        # TODO Walk node to determine size before handing off to serializer.
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                entry.structure_family,
                serialization_registry,
                data,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
                filter_for_access=curried_filter,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/awkward/buffers/{path:path}",
    response_model=schemas.Response,
    name="AwkwardArray buffers",
)
async def get_awkward_buffers(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    form_key: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a slice of AwkwardArray data.

    Note that there is a POST route on this same path with equivalent functionality.
    HTTP caches tends to engage with GET but not POST, so that GET route may be
    preferred for that reason. However, HTTP clients, servers, and proxies
    typically impose a length limit on URLs. (The HTTP spec does not specify
    one, but this is a pragmatic measure.) For requests with large numbers of
    form_key parameters, POST may be the only option.
    """
    return await _awkward_buffers(
        request=request,
        entry=entry,
        form_key=form_key,
        format=format,
        filename=filename,
        serialization_registry=serialization_registry,
        settings=settings,
    )


@router.post(
    "/awkward/buffers/{path:path}",
    response_model=schemas.Response,
    name="AwkwardArray buffers",
)
async def post_awkward_buffers(
    request: Request,
    body: List[str],
    entry=SecureEntry(scopes=["read:data"]),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a slice of AwkwardArray data.

    Note that there is a GET route on this same path with equivalent functionality.
    HTTP caches tends to engage with GET but not POST, so that GET route may be
    preferred for that reason. However, HTTP clients, servers, and proxies
    typically impose a length limit on URLs. (The HTTP spec does not specify
    one, but this is a pragmatic measure.) For requests with large numbers of
    form_key parameters, POST may be the only option.
    """
    return await _awkward_buffers(
        request=request,
        entry=entry,
        form_key=body,
        format=format,
        filename=filename,
        serialization_registry=serialization_registry,
        settings=settings,
    )


async def _awkward_buffers(
    request: Request,
    entry,
    form_key: Optional[List[str]],
    format: Optional[str],
    filename: Optional[str],
    serialization_registry,
    settings: BaseSettings,
):
    structure_family = entry.structure_family
    structure = entry.structure()
    if structure_family != StructureFamily.awkward:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /awkward/buffers route.",
        )
    with record_timing(request.state.metrics, "read"):
        # The plural vs. singular mismatch is due to the way query parameters
        # are given as ?form_key=A&form_key=B&form_key=C.
        container = await ensure_awaitable(entry.read_buffers, form_key)
    if (
        sum(len(buffer) for buffer in container.values())
        > settings.response_bytesize_limit
    ):
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    components = (structure.form, structure.length, container)
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                structure_family,
                serialization_registry,
                components,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/awkward/full/{path:path}",
    response_model=schemas.Response,
    name="Full AwkwardArray",
)
async def awkward_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    # slice=Depends(slice_),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a slice of AwkwardArray data.
    """
    structure_family = entry.structure_family
    if structure_family != StructureFamily.awkward:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /awkward/full route.",
        )
    # Deferred import because this is not a required dependency of the server
    # for some use cases.
    import awkward

    with record_timing(request.state.metrics, "read"):
        container = await ensure_awaitable(entry.read)
    structure = entry.structure()
    components = (structure.form, structure.length, container)
    array = awkward.from_buffers(*components)
    if array.nbytes > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return await construct_data_response(
                structure_family,
                serialization_registry,
                components,
                entry.metadata(),
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.post("/metadata/{path:path}", response_model=schemas.PostMetadataResponse)
async def post_metadata(
    request: Request,
    path: str,
    body: schemas.PostMetadataRequest,
    validation_registry=Depends(get_validation_registry),
    settings: BaseSettings = Depends(get_settings),
    entry=SecureEntry(scopes=["write:metadata", "create"]),
):
    if not getattr(entry, "writable", False):
        raise HTTPException(
            status_code=405, detail=f"Data cannot be written at the path {path}"
        )

    metadata, structure_family, specs = (
        body.metadata,
        body.structure_family,
        body.specs,
    )
    if structure_family == StructureFamily.container:
        structure = None
    else:
        if len(body.data_sources) != 1:
            raise NotImplementedError
        structure = body.data_sources[0].structure

    metadata_modified = False

    # Specs should be ordered from most specific/constrained to least.
    # Validate them in reverse order, with the least constrained spec first,
    # because it may do normalization that helps pass the more constrained one.
    # Known Issue:
    # When there is more than one spec, it's possible for the validator for
    # Spec 2 to make a modification that breaks the validation for Spec 1.
    # For now we leave it to the server maintainer to ensure that validators
    # won't step on each other in this way, but this may need revisiting.
    for spec in reversed(specs):
        if spec.name not in validation_registry:
            if settings.reject_undeclared_specs:
                raise HTTPException(
                    status_code=400, detail=f"Unrecognized spec: {spec.name}"
                )
        else:
            validator = validation_registry(spec.name)
            try:
                result = validator(metadata, structure_family, structure, spec)
            except ValidationError as e:
                raise HTTPException(
                    status_code=400,
                    detail=f"failed validation for spec {spec.name}:\n{e}",
                )
            if result is not None:
                metadata_modified = True
                metadata = result

    key, node = await entry.create_node(
        metadata=body.metadata,
        structure_family=body.structure_family,
        key=body.id,
        specs=body.specs,
        data_sources=body.data_sources,
    )
    links = {}
    base_url = get_base_url(request)
    path_parts = [segment for segment in path.split("/") if segment] + [key]
    path_str = "/".join(path_parts)
    links["self"] = f"{base_url}/metadata/{path_str}"
    if body.structure_family in {StructureFamily.array, StructureFamily.sparse}:
        block_template = ",".join(
            f"{{{index}}}" for index in range(len(node.structure().shape))
        )
        links["block"] = f"{base_url}/array/block/{path_str}?block={block_template}"
        links["full"] = f"{base_url}/array/full/{path_str}"
    elif body.structure_family == StructureFamily.table:
        links[
            "partition"
        ] = f"{base_url}/table/partition/{path_str}?partition={{index}}"
        links["full"] = f"{base_url}/table/full/{path_str}"
    elif body.structure_family == StructureFamily.container:
        links["full"] = f"{base_url}/container/full/{path_str}"
        links["search"] = f"{base_url}/search/{path_str}"
    elif body.structure_family == StructureFamily.awkward:
        links["buffers"] = f"{base_url}/awkward/buffers/{path_str}"
        links["full"] = f"{base_url}/awkward/full/{path_str}"
    else:
        raise NotImplementedError(body.structure_family)
    response_data = {
        "id": key,
        "links": links,
        "data_sources": [ds.dict() for ds in node.data_sources],
    }
    if metadata_modified:
        response_data["metadata"] = metadata
    return json_or_msgpack(request, response_data)


@router.delete("/metadata/{path:path}")
async def delete(
    request: Request,
    entry=SecureEntry(scopes=["write:data", "write:metadata"]),
):
    if hasattr(entry, "delete"):
        await entry.delete()
    else:
        raise HTTPException(
            status_code=405, detail="This node does not support deletion."
        )
    return json_or_msgpack(request, None)


@router.put("/array/full/{path:path}")
async def put_array_full(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    deserialization_registry=Depends(get_deserialization_registry),
):
    body = await request.body()
    if not hasattr(entry, "write"):
        raise HTTPException(
            status_code=405, detail="This node cannot accept array data."
        )
    media_type = request.headers["content-type"]
    if entry.structure_family == "array":
        dtype = entry.structure().data_type.to_numpy_dtype()
        shape = entry.structure().shape
        deserializer = deserialization_registry.dispatch("array", media_type)
        data = await ensure_awaitable(deserializer, body, dtype, shape)
    elif entry.structure_family == "sparse":
        deserializer = deserialization_registry.dispatch("sparse", media_type)
        data = await ensure_awaitable(deserializer, body)
    else:
        raise NotImplementedError(entry.structure_family)
    await ensure_awaitable(entry.write, data)
    return json_or_msgpack(request, None)


@router.put("/array/block/{path:path}")
async def put_array_block(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    deserialization_registry=Depends(get_deserialization_registry),
    block=Depends(block),
):
    if not hasattr(entry, "write_block"):
        raise HTTPException(
            status_code=405, detail="This node cannot accept array data."
        )
    from tiled.adapters.array import slice_and_shape_from_block_and_chunks

    body = await request.body()
    media_type = request.headers["content-type"]
    if entry.structure_family == "array":
        dtype = entry.structure().data_type.to_numpy_dtype()
        _, shape = slice_and_shape_from_block_and_chunks(
            block, entry.structure().chunks
        )
        deserializer = deserialization_registry.dispatch("array", media_type)
        data = await ensure_awaitable(deserializer, body, dtype, shape)
    elif entry.structure_family == "sparse":
        deserializer = deserialization_registry.dispatch("sparse", media_type)
        data = await ensure_awaitable(deserializer, body)
    else:
        raise NotImplementedError(entry.structure_family)
    await ensure_awaitable(entry.write_block, data, block)
    return json_or_msgpack(request, None)


@router.put("/table/full/{path:path}")
@router.put("/node/full/{path:path}", deprecated=True)
async def put_node_full(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    deserialization_registry=Depends(get_deserialization_registry),
):
    if not hasattr(entry, "write"):
        raise HTTPException(
            status_code=405, detail="This node does not support writing."
        )
    body = await request.body()
    media_type = request.headers["content-type"]
    deserializer = deserialization_registry.dispatch(StructureFamily.table, media_type)
    data = await ensure_awaitable(deserializer, body)
    await ensure_awaitable(entry.write, data)
    return json_or_msgpack(request, None)


@router.put("/table/partition/{path:path}")
async def put_table_partition(
    partition: int,
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    deserialization_registry=Depends(get_deserialization_registry),
):
    if not hasattr(entry, "write_partition"):
        raise HTTPException(
            status_code=405, detail="This node does not supporting writing a partition."
        )
    body = await request.body()
    media_type = request.headers["content-type"]
    deserializer = deserialization_registry.dispatch(StructureFamily.table, media_type)
    data = await ensure_awaitable(deserializer, body)
    await ensure_awaitable(entry.write_partition, data, partition)
    return json_or_msgpack(request, None)


@router.put("/awkward/full/{path:path}")
async def put_awkward_full(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    deserialization_registry=Depends(get_deserialization_registry),
):
    body = await request.body()
    if entry.structure_family != StructureFamily.awkward:
        raise HTTPException(
            status_code=404, detail="This route is not applicable to this node."
        )
    if not hasattr(entry, "write"):
        raise HTTPException(status_code=405, detail="This node cannot be written to.")
    media_type = request.headers["content-type"]
    deserializer = deserialization_registry.dispatch(
        StructureFamily.awkward, media_type
    )
    structure = entry.structure()
    data = await ensure_awaitable(deserializer, body, structure.form, structure.length)
    await ensure_awaitable(entry.write, data)
    return json_or_msgpack(request, None)


@router.put("/metadata/{path:path}", response_model=schemas.PutMetadataResponse)
async def put_metadata(
    request: Request,
    body: schemas.PutMetadataRequest,
    validation_registry=Depends(get_validation_registry),
    settings: BaseSettings = Depends(get_settings),
    entry=SecureEntry(scopes=["write:metadata"]),
):
    if not hasattr(entry, "update_metadata"):
        raise HTTPException(
            status_code=405, detail="This node does not support update of metadata."
        )

    metadata, structure_family, structure, specs = (
        body.metadata if body.metadata is not None else entry.metadata(),
        entry.structure_family,
        entry.structure(),
        body.specs if body.specs is not None else entry.specs,
    )

    metadata_modified = False

    # Specs should be ordered from most specific/constrained to least.
    # Validate them in reverse order, with the least constrained spec first,
    # because it may do normalization that helps pass the more constrained one.
    # Known Issue:
    # When there is more than one spec, it's possible for the validator for
    # Spec 2 to make a modification that breaks the validation for Spec 1.
    # For now we leave it to the server maintainer to ensure that validators
    # won't step on each other in this way, but this may need revisiting.
    for spec in reversed(specs):
        if spec.name not in validation_registry:
            if settings.reject_undeclared_specs:
                raise HTTPException(
                    status_code=400, detail=f"Unrecognized spec: {spec.name}"
                )
        else:
            validator = validation_registry(spec.name)
            try:
                result = validator(metadata, structure_family, structure, spec)
            except ValidationError as e:
                raise HTTPException(
                    status_code=400,
                    detail=f"failed validation for spec {spec.name}:\n{e}",
                )
            if result is not None:
                metadata_modified = True
                metadata = result

    await entry.update_metadata(metadata=metadata, specs=specs)

    response_data = {"id": entry.key}
    if metadata_modified:
        response_data["metadata"] = metadata
    return json_or_msgpack(request, response_data)


@router.get("/revisions/{path:path}")
async def get_revisions(
    request: Request,
    path: str,
    offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
    limit: Optional[int] = Query(
        DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
    ),
    entry=SecureEntry(scopes=["read:metadata"]),
):
    if not hasattr(entry, "revisions"):
        raise HTTPException(
            status_code=405, detail="This node does not support revisions."
        )

    base_url = get_base_url(request)
    resource = await construct_revisions_response(
        entry,
        base_url,
        "/revisions",
        path,
        offset,
        limit,
        resolve_media_type(request),
    )
    return json_or_msgpack(request, resource.dict())


@router.delete("/revisions/{path:path}")
async def delete_revision(
    request: Request,
    number: int,
    entry=SecureEntry(scopes=["write:metadata"]),
):
    if not hasattr(entry, "revisions"):
        raise HTTPException(
            status_code=405,
            detail="This node does not support a del request for revisions.",
        )

    await entry.delete_revision(number)
    return json_or_msgpack(request, None)



from dataclasses import dataclass

import awkward


@dataclass
class AwkwardStructure:
    length: int
    form: dict

    @classmethod
    def from_json(cls, structure):
        return cls(**structure)


def project_form(form, form_keys_touched):
    # See https://github.com/bluesky/tiled/issues/450
    if isinstance(form, awkward.forms.RecordForm):
        if form.fields is None:
            original_fields = [None] * len(form.contents)
        else:
            original_fields = form.fields

        fields = []
        contents = []
        for field, content in zip(original_fields, form.contents):
            projected = project_form(content, form_keys_touched)
            if projected is not None:
                fields.append(field)
                contents.append(projected)

        if form.fields is None:
            fields = None

        return form.copy(fields=fields, contents=contents)

    elif isinstance(form, awkward.forms.UnionForm):
        step1 = [project_form(x, form_keys_touched) for x in form.contents]
        step2 = [x for x in step1 if x is not None]
        if len(step2) == 0:
            return None
        elif len(step2) == 1:
            return step2[0]
        elif step2 == form.contents:
            return form
        else:
            raise NotImplementedError(
                "Certain UnionForms are not yet supported. "
                "See https://github.com/scikit-hep/awkward/issues/2666"
            )
            return awkward.forms.UnionForm.simplified(
                form.tags,
                form.index,
                step2,
                parameters=form.parameters,
                form_key=form.form_key,
            )

    elif isinstance(form, (awkward.forms.NumpyForm, awkward.forms.EmptyForm)):
        if form.form_key in form_keys_touched:
            return form.copy()
        else:
            return None

    elif isinstance(form, (awkward.forms.RegularForm, awkward.forms.UnmaskedForm)):
        return form.copy(content=project_form(form.content, form_keys_touched))

    else:
        if form.form_key in form_keys_touched:
            return form.copy(content=project_form(form.content, form_keys_touched))
        else:
            return None



# There is no public API in httpx to injecting additional decoders.
from httpx._decoders import SUPPORTED_DECODERS

from ..utils import modules_available

if modules_available("blosc"):

    class BloscDecoder:
        def __init__(self):
            # Blosc seems to have no streaming interface.
            # Accumulate response data in a cache here,
            # and concatenate and decode at the end.
            self._data = []

        def decode(self, data: bytes) -> bytes:
            self._data.append(data)
            return b""

        def flush(self) -> bytes:
            # Hide this here to defer the numpy import that it triggers.
            import blosc

            if len(self._data) == 1:
                (data,) = self._data
            else:
                data = b"".join(self._data)
            return blosc.decompress(data)

    SUPPORTED_DECODERS["blosc"] = BloscDecoder


if modules_available("zstandard"):
    import zstandard

    class ZStandardDecoder:
        def __init__(self):
            self._context = zstandard.ZstdDecompressor()
            self._obj = self._context.decompressobj()

        def decode(self, data: bytes) -> bytes:
            return self._obj.decompress(data)

        def flush(self) -> bytes:
            return b""

    SUPPORTED_DECODERS["zstd"] = ZStandardDecoder



import contextlib
import os

import numpy
import pandas
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app_from_config
from .utils import fail_with_status_code

tiny_array = numpy.ones(5)
tiny_df = pandas.DataFrame({"a": tiny_array})
small_array = numpy.ones(50)
small_df = pandas.DataFrame({"a": small_array})
size_limit = small_array.nbytes / 2
assert tiny_array.nbytes < size_limit < small_array.nbytes
assert tiny_df.memory_usage().sum() < size_limit < small_df.memory_usage().sum()


tree = MapAdapter(
    {
        "tiny_array": ArrayAdapter.from_array(tiny_array),
        "small_array": ArrayAdapter.from_array(small_array),
        "tiny_df": DataFrameAdapter.from_pandas(tiny_df, npartitions=1),
        "small_df": DataFrameAdapter.from_pandas(small_df, npartitions=2),
    }
)
config = {
    "trees": [
        {
            "tree": f"{__name__}:tree",
            "path": "/",
        },
    ],
    "response_bytesize_limit": size_limit,
}


@pytest.fixture(scope="module")
def client():
    app = build_app_from_config(config)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@contextlib.contextmanager
def low_size_limit():
    os.environ["TILED_RESPONSE_BYTESIZE_LIMIT"] = str(size_limit)
    yield
    os.environ.pop("TILED_RESPONSE_BYTESIZE_LIMIT")


def test_array(client, tmpdir):
    """
    Download an array over the size limit.
    """
    with low_size_limit():
        path = str(tmpdir / "test.csv")
        client["tiny_array"].read()  # This is fine.
        client["tiny_array"].export(path)  # This is fine.
        with fail_with_status_code(400):
            client["small_array"].read()  # too big
        with fail_with_status_code(400):
            client["small_array"].export(path)  # too big


def test_dataframe(client, tmpdir):
    """
    Download an dataframe over the size limit.
    """

    with low_size_limit():
        path = str(tmpdir / "test.csv")
        client["tiny_df"].read()  # This is fine.
        client["tiny_df"].export(path)
        with fail_with_status_code(400):
            client["small_df"].read()  # too big
        with fail_with_status_code(400):
            client["small_df"].export(path)  # too big



"""
These objects describe queries, independent of how the actual querying is implemented.
They are used on the client side and the server side.

The are encoded into and decoded from URL query parameters.
"""

import enum
import json
from dataclasses import dataclass
from typing import Any, List

from .query_registration import register
from .structures.core import StructureFamily as StructureFamilyEnum

JSONSerializable = Any  # Feel free to refine this.


class NoBool:
    def __bool__(self):
        raise TypeError(
            """Queries are not "truth-y" or "false-y". They must be passed to search().

You may be seeing this error message because you tried to use a tiled query
with the Python keywords `and` or `or`. This does not (cannot) work.
To compose queries, chain search calls like:

    c.search(...).search(...).search(...)
"""
        )


@register(name="fulltext")
@dataclass
class FullText(NoBool):
    """
    Search the full text of all metadata values for word matches.

    This matches *complete words*, so 'dog' would match 'cat dog elephant',
    but 'do' would not match.

    Parameters
    ----------
    text : str
    """

    text: str

    def encode(self):
        return {"text": self.text}

    @classmethod
    def decode(cls, *, text):
        return cls(text=text)


@register(name="lookup")
@dataclass
class KeyLookup(NoBool):
    """
    Match a specific Entry by key. Mostly for internal use.

    This is necessary to support item lookup within search results, as in:

    >>> tree.search(...)["..."]

    The server handles this directly and generically, simply calling __getitem__
    on the tree after apply all other queries.  Implementations of search(...)
    do not need to handle it.

    Parameters
    ----------
    key : str
    """

    key: str

    def encode(self):
        return {"key": self.key}

    @classmethod
    def decode(cls, *, key):
        return cls(key=key)


@register(name="keys_filter")
@dataclass
class KeysFilter(NoBool):
    """
    Filter entries that do not match one of these keys.

    This is used by the SimpleAccessPolicy.

    Parameters
    ----------
    keys : List[str]
    """

    keys: List[str]

    def encode(self):
        return {"keys": json.dumps(self.keys)}

    @classmethod
    def decode(cls, *, keys):
        return cls(keys=json.loads(keys))


@register(name="regex")
@dataclass
class Regex(NoBool):
    """
    Match a key's value to a regular expression.

    Parameters
    ----------
    key : str
        e.g. "color", "sample.name"
    pattern : str
        regular expression
    case_sensitive : bool, optional
        Default True (case-sensitive).
        Note that this is the opposite of the default for FullText;
        regex users generally expect case sensitivity by default.

    Examples
    --------

    Search for color == "red"

    >>> c.search(Regex("sample.name", "Cu.*"))
    """

    key: str
    pattern: str
    case_sensitive: bool = True

    def encode(self):
        return {
            "key": self.key,
            "pattern": self.pattern,
            "case_sensitive": json.dumps(self.case_sensitive),
        }

    @classmethod
    def decode(cls, *, key, pattern, case_sensitive=True):
        # Note: FastAPI decodes case_sensitive into a boolean for us.
        return cls(
            key=key,
            pattern=pattern,
            case_sensitive=case_sensitive,
        )


@register(name="eq")
@dataclass
class Eq(NoBool):
    """
    Query equality of a given key's value to the specified value.

    See `Key` in this module for a more intuitive interface for equality.

    Parameters
    ----------
    key : str
        e.g. "color", "sample.name"
    value : JSONSerializable
        May be a string, number, list, or dict.

    Examples
    --------

    Search for color == "red"

    >>> c.search(Eq("color", "red"))
    """

    key: str
    value: JSONSerializable

    def encode(self):
        return {"key": self.key, "value": json.dumps(self.value)}

    @classmethod
    def decode(cls, *, key, value):
        return cls(key=key, value=json.loads(value))


@register(name="noteq")
@dataclass
class NotEq(NoBool):
    """
    Query inequality of a given key's value to the specified value.

    See `Key` in this module for a more intuitive interface for inequality.

    Parameters
    ----------
    key : str
        e.g. "color", "sample.name"
    value : JSONSerializable
        May be a string, number, list, or dict.

    Examples
    --------

    Search for color == "red"

    >>> c.search(Eq("color", "red"))
    """

    key: str
    value: JSONSerializable

    def encode(self):
        return {"key": self.key, "value": json.dumps(self.value)}

    @classmethod
    def decode(cls, *, key, value):
        return cls(key=key, value=json.loads(value))


class Operator(str, enum.Enum):
    lt = "lt"
    gt = "gt"
    le = "le"
    ge = "ge"


@register(name="comparison")
@dataclass
class Comparison(NoBool):
    """
    Query binary comparison between given key's value to the specified value.

    See `Key` in this module for a more intuitive interface for comparisons.

    Parameters
    ----------
    operator : {"gt", "lt", "ge", "lt"}
    key : str
        e.g. "temperature"
    value : JSONSerializable
        May be a string, number, list, or dict.

    Examples
    --------

    Search for temperature > 300.

    >>> c.search(Comparison("gt", "temperature", 300))
    """

    operator: Operator
    key: str
    value: JSONSerializable

    def __init__(self, operator, key, value):
        self.operator = Operator(operator)
        self.key = key
        self.value = value

    def encode(self):
        return {
            "operator": self.operator.value,
            "key": self.key,
            "value": json.dumps(self.value),
        }

    @classmethod
    def decode(cls, *, operator, key, value):
        return cls(operator=Operator(operator), key=key, value=json.loads(value))


@register(name="contains")
@dataclass
class Contains(NoBool):
    """
    Query where a given key's value contains the specified value.

    Parameters
    ----------
    key : str
        e.g. "motors"
    value : JSONSerializable
        May be a string, number, list, or dict.

    Examples
    --------

    Search for matches where "ccd" is including the list of detectors.

    >>> c.search(Contains("detectors", "ccd"))
    """

    key: str
    value: JSONSerializable

    def encode(self):
        return {"key": self.key, "value": json.dumps(self.value)}

    @classmethod
    def decode(cls, *, key, value):
        return cls(key=key, value=json.loads(value))


@register(name="in")
@dataclass
class In:
    """
    Query if a given key's value is present in the specified sequence of values.

    Parameters
    ----------
    key : str
        e.g. "color", "sample.name"
    value : List[JSONSerializable]
        e.g. ["red", "blue"]

    Examples
    --------

    Search for color in ["red", "blue"]

    >>> c.search(In("color", ["red", "blue"]))
    """

    key: str
    value: List[JSONSerializable]

    def __post_init__(self):
        self.value = list(self.value)

    def encode(self):
        return {"key": self.key, "value": json.dumps(self.value)}

    @classmethod
    def decode(cls, *, key, value):
        return cls(key=key, value=json.loads(value))


@register(name="notin")
@dataclass
class NotIn:
    """
    Query if a given key's value is not present in the specified sequence of values.

    Parameters
    ----------
    key : str
        e.g. "color", "sample.name"
    value : List[JSONSerializable]
        e.g. ["red", "blue"]

    Examples
    --------

    Search for color not in ["red", "blue"]

    >>> c.search(NotIn("color", ["red", "blue"]))
    """

    key: str
    value: List[JSONSerializable]

    def __post_init__(self):
        self.value = list(self.value)

    def encode(self):
        return {"key": self.key, "value": json.dumps(self.value)}

    @classmethod
    def decode(cls, *, key, value):
        return cls(key=key, value=json.loads(value))


@register(name="specs")
@dataclass(init=False)
class SpecsQuery:
    """
    Query if specs list matches all elements in include list and does not match any element in exclude list

    Parameters
    ----------
    include : List[str]
    exclude : List[str]

    Examples
    --------

    Search for specs ["foo", "bar"] and NOT "baz"

    >>> c.search(SpecsQuery(include=["foo", "bar"], exclude=["baz"]))
    """

    include: List[str]
    exclude: List[str]

    def __init__(self, include, exclude=None):
        exclude = exclude or []

        if isinstance(include, str):
            raise TypeError("include must be a list not a str")

        if isinstance(exclude, str):
            raise TypeError("exclude must be a list not a str")

        self.include = list(include)
        self.exclude = list(exclude)

    def encode(self):
        return {
            "include": json.dumps(self.include),
            "exclude": json.dumps(self.exclude),
        }

    @classmethod
    def decode(cls, *, include, exclude):
        return cls(include=json.loads(include), exclude=json.loads(exclude))


def SpecQuery(spec):
    """
    Convenience function for querying if specs list contains a given spec

    Equivalent to SpecsQuery([spec]).

    Parameters
    ----------
    spec: str

    Examples
    --------

    Search for spec "foo"

    >>> c.search(Spec("foo"))
    """

    return SpecsQuery([spec])


@register(name="structure_family")
@dataclass(init=False)
class StructureFamilyQuery:
    """
    Query if structure_families match value

    Parameters
    ----------
    value : StructureFamily

    Examples
    --------

    Search for dataframes

    >>> c.search(StructureFamilies("dataframe"))
    """

    def __init__(self, value):
        self.value = StructureFamilyEnum(value)

    value: StructureFamilyEnum

    def encode(self):
        return {"value": self.value.value}

    @classmethod
    def decode(cls, *, value):
        return cls(value=value)


class Key:
    """
    Compare a key in the metadata to a value using standard Python operators.

    This itself is not a query, but *comparing* it with a value, as shown
    in the examples below, produces a query.

    Parameters
    ----------
    key : str

    Examples
    --------

    Search for equality, comparison, or membership in a collection.

    >>> c.search(Key("color") == "red")
    >>> c.search(Key("temperature") >= 300)
    >>> c.search(Key("temperature") <= 300)
    >>> c.search(Key("position") > 5.0)
    >>> c.search(Key("position") < 5.0)
    """

    def __init__(self, key):
        self.key = key

    def __eq__(self, value):
        return Eq(self.key, value)

    def __ne__(self, value):
        return NotEq(self.key, value)

    def __lt__(self, value):
        return Comparison("lt", self.key, value)

    def __gt__(self, value):
        return Comparison("gt", self.key, value)

    def __le__(self, value):
        return Comparison("le", self.key, value)

    def __ge__(self, value):
        return Comparison("ge", self.key, value)

    # Note: __contains__ cannot be supported because the language coerces
    # the result of __contains__ to be a literal boolean. We are not
    # allowed to return a custom type.


class QueryValueError(ValueError):
    pass



import io
import mimetypes

from ..media_type_registration import deserialization_registry, serialization_registry
from ..structures.core import StructureFamily
from ..utils import APACHE_ARROW_FILE_MIME_TYPE, XLSX_MIME_TYPE, modules_available


@serialization_registry.register(StructureFamily.table, APACHE_ARROW_FILE_MIME_TYPE)
def serialize_arrow(df, metadata, preserve_index=True):
    import pyarrow

    table = pyarrow.Table.from_pandas(df, preserve_index=preserve_index)
    sink = pyarrow.BufferOutputStream()
    with pyarrow.ipc.new_file(sink, table.schema) as writer:
        writer.write_table(table)
    return memoryview(sink.getvalue())


@deserialization_registry.register(StructureFamily.table, APACHE_ARROW_FILE_MIME_TYPE)
def deserialize_arrow(buffer):
    import pyarrow

    return pyarrow.ipc.open_file(buffer).read_pandas()


# There seems to be no official Parquet MIME type.
# https://issues.apache.org/jira/browse/PARQUET-1889
@serialization_registry.register(StructureFamily.table, "application/x-parquet")
def serialize_parquet(df, metadata, preserve_index=True):
    import pyarrow.parquet

    table = pyarrow.Table.from_pandas(df, preserve_index=preserve_index)
    sink = pyarrow.BufferOutputStream()
    with pyarrow.parquet.ParquetWriter(sink, table.schema) as writer:
        writer.write_table(table)
    return memoryview(sink.getvalue())


def serialize_csv(df, metadata, preserve_index=False):
    file = io.StringIO()
    df.to_csv(file, index=preserve_index)
    return file.getvalue().encode()


serialization_registry.register(StructureFamily.table, "text/csv", serialize_csv)
serialization_registry.register(
    StructureFamily.table, "text/x-comma-separated-values", serialize_csv
)
serialization_registry.register(StructureFamily.table, "text/plain", serialize_csv)


@serialization_registry.register(StructureFamily.table, "text/html")
def serialize_html(df, metadata, preserve_index=False):
    file = io.StringIO()
    df.to_html(file, index=preserve_index)
    return file.getvalue().encode()


if modules_available("openpyxl", "pandas"):
    # The optional pandas dependency openpyxel is required for Excel read/write.
    import pandas

    @serialization_registry.register(StructureFamily.table, XLSX_MIME_TYPE)
    def serialize_excel(df, metadata, preserve_index=False):
        file = io.BytesIO()
        df.to_excel(file, index=preserve_index)
        return file.getbuffer()

    deserialization_registry.register(
        StructureFamily.table, XLSX_MIME_TYPE, pandas.read_excel
    )
    mimetypes.types_map.setdefault(".xlsx", XLSX_MIME_TYPE)
if modules_available("orjson"):
    import orjson

    serialization_registry.register(
        StructureFamily.table,
        "application/json",
        lambda df, metadata: orjson.dumps(
            {column: df[column].tolist() for column in df},
        ),
    )

    # Newline-delimited JSON. For example, this DataFrame:
    #
    # >>> pandas.DataFrame({"a": [1,2,3], "b": [4,5,6]})
    #
    # renders as this multi-line output:
    #
    # {'a': 1, 'b': 4}
    # {'a': 2, 'b': 5}
    # {'a': 3, 'b': 6}
    @serialization_registry.register(
        StructureFamily.table,
        "application/json-seq",  # official mimetype for newline-delimited JSON
    )
    def json_sequence(df, metadata):
        rows = df.iterrows()
        # The first row is a special case; the rest start with a newline.
        try:
            # Emit the first row with no newline.
            _, row = next(rows)
        except StopIteration:
            # No rows
            yield b""
        else:
            # Emit the remaining rows, prepending newline.
            yield orjson.dumps(row.to_dict())
            for _, row in rows:
                yield b"\n" + orjson.dumps(row.to_dict())


if modules_available("h5py"):
    from .container import serialize_hdf5

    serialization_registry.register(
        StructureFamily.table, "application/x-hdf5", serialize_hdf5
    )



import yaml

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..config import parse_configs
from ..server.app import build_app_from_config

tree = MapAdapter({"example": ArrayAdapter.from_array([1, 2, 3])})


def test_root():
    "One tree served at top level"
    config = {
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/",
            },
        ]
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        assert list(client) == ["example"]


def test_single_nested():
    "One tree served nested one layer down"
    config = {
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/a/b",
            },
        ]
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        assert list(client) == ["a"]
        assert list(client["a"]) == ["b"]
        assert list(client["a"]["b"]) == ["example"]


def test_single_deeply_nested():
    "One tree served nested many layers down"
    config = {
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/a/b/c/d/e",
            },
        ]
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        assert list(client) == ["a"]
        assert list(client["a"]) == ["b"]
        assert list(client["a"]["b"]) == ["c"]
        assert list(client["a"]["b"]["c"]) == ["d"]
        assert list(client["a"]["b"]["c"]["d"]) == ["e"]
        assert list(client["a"]["b"]["c"]["d"]["e"]) == ["example"]


def test_many_nested():
    config = {
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/a/b",
            },
            {
                "tree": f"{__name__}:tree",
                "path": "/a/c",
            },
            {
                "tree": f"{__name__}:tree",
                "path": "/a/d/e",
            },
            {
                "tree": f"{__name__}:tree",
                "path": "/a/d/f",
            },
            {
                "tree": f"{__name__}:tree",
                "path": "/a/d/g/h",
            },
            {
                "tree": f"{__name__}:tree",
                "path": "/a/d/g/i",
            },
        ],
    }
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        assert list(client["a"]["b"]) == ["example"]
        assert list(client["a"]["c"]) == ["example"]
        assert list(client["a"]["d"]["e"]) == ["example"]
        assert list(client["a"]["d"]["f"]) == ["example"]
        assert list(client["a"]["d"]["g"]["h"]) == ["example"]
        assert list(client["a"]["d"]["g"]["i"]) == ["example"]


def test_extra_files(tmpdir):
    config = {"trees": [{"path": "/", "tree": "tiled.examples.generated_minimal:tree"}]}
    with open(tmpdir / "config.yml", "w") as config_file:
        yaml.dump(config, config_file)
    with open(tmpdir / "README.md", "w") as extra_file:
        extra_file.write("# Example")
    parse_configs(str(tmpdir))



"""
The ArrayStructure class in tiled.structure.array is implemeneted with Python
built-in dataclasses. This is an implementation of the same structure in pydantic.

FastAPI requires pydantic models specifically, and will attempt to automatically
convert dataclasses to pydantic models. This normally works fine, but in the
case of ArrayStructure specifically it fails because of this issue:

    https://github.com/samuelcolvin/pydantic/issues/3251

Once this problem is fixed in pydantic, this separate implementation in
ArrayStructure will no longer by needed. We will turn this module into a shim
that just imports ArrayStructure from tiled.structures.array and then deprecate
it.
"""

import sys
from typing import List, Optional, Tuple, Union

import numpy
from pydantic import BaseModel

from ..structures.array import Endianness, Kind


class BuiltinDtype(BaseModel):
    endianness: Endianness
    kind: Kind
    itemsize: int

    __endianness_map = {
        ">": "big",
        "<": "little",
        "=": sys.byteorder,
        "|": "not_applicable",
    }

    __endianness_reverse_map = {"big": ">", "little": "<", "not_applicable": "|"}

    @classmethod
    def from_numpy_dtype(cls, dtype):
        return cls(
            endianness=cls.__endianness_map[dtype.byteorder],
            kind=Kind(dtype.kind),
            itemsize=dtype.itemsize,
        )

    def to_numpy_dtype(self):
        return numpy.dtype(self.to_numpy_str())

    def to_numpy_str(self):
        endianness = self.__endianness_reverse_map[self.endianness]
        # dtype.itemsize always reports bytes.  The format string from the
        # numeric types the string format is: {type_code}{byte_count} so we can
        # directly use the item size.
        #
        # for unicode the pattern is 'U{char_count}', however
        # which numpy always represents as 4 byte UCS4 encoding
        # (because variable width encodings do not fit with fixed-stride arrays)
        # so the reported itemsize is 4x the char count.  To get back to the string
        # we need to divide by 4.
        size = self.itemsize if self.kind != Kind.unicode else self.itemsize // 4
        return f"{endianness}{self.kind.value}{size}"

    @classmethod
    def from_json(cls, structure):
        return cls(
            kind=Kind(structure["kind"]),
            itemsize=structure["itemsize"],
            endianness=Endianness(structure["endianness"]),
        )


class Field(BaseModel):
    name: str
    dtype: Union[BuiltinDtype, "StructDtype"]
    shape: Optional[Tuple[int, ...]]

    @classmethod
    def from_numpy_descr(cls, field):
        name, *rest = field
        if name == "":
            raise ValueError(
                f"You seem to have gotten descr of a base or subdtype: {field}"
            )
        if len(rest) == 1:
            (f_type,) = rest
            shape = None
        else:
            f_type, shape = rest

        if isinstance(f_type, str):
            FType = BuiltinDtype.from_numpy_dtype(numpy.dtype(f_type))
        else:
            FType = StructDtype.from_numpy_dtype(numpy.dtype(f_type))
        return cls(name=name, dtype=FType, shape=shape)

    def to_numpy_descr(self):
        if isinstance(self.dtype, BuiltinDtype):
            base = [self.name, self.dtype.to_numpy_str()]
        else:
            base = [self.name, self.dtype.to_numpy_descr()]
        if self.shape is None:
            return tuple(base)
        else:
            return tuple(base + [self.shape])

    @classmethod
    def from_json(cls, structure):
        name = structure["name"]
        if "fields" in structure["dtype"]:
            ftype = StructDtype.from_json(structure["dtype"])
        else:
            ftype = BuiltinDtype.from_json(structure["dtype"])
        return cls(name=name, dtype=ftype, shape=structure["shape"])


class StructDtype(BaseModel):
    itemsize: int
    fields: List[Field]

    @classmethod
    def from_numpy_dtype(cls, dtype):
        # subdtypes push extra dimensions into arrays, we should handle these
        # a layer up and report an array with bigger dimensions.
        if dtype.subdtype is not None:
            raise ValueError(f"We do not know how to encode subdtypes: {dtype}")
        # If this is a builtin type, require the use of BuiltinDtype (nee .array.BuiltinDtype)
        if dtype.fields is None:
            raise ValueError(f"You have a base type: {dtype}")
        return cls(
            itemsize=dtype.itemsize,
            fields=[Field.from_numpy_descr(f) for f in dtype.descr],
        )

    def to_numpy_dtype(self):
        return numpy.dtype(self.to_numpy_descr())

    def to_numpy_descr(self):
        return [f.to_numpy_descr() for f in self.fields]

    def max_depth(self):
        return max(
            1 if isinstance(f.dtype, BuiltinDtype) else 1 + f.dtype.max_depth()
            for f in self.fields
        )

    @classmethod
    def from_json(cls, structure):
        return cls(
            itemsize=structure["itemsize"],
            fields=[Field.from_json(f) for f in structure["fields"]],
        )


Field.update_forward_refs()


class ArrayStructure(BaseModel):
    data_type: Union[BuiltinDtype, StructDtype]
    chunks: Tuple[Tuple[int, ...], ...]  # tuple-of-tuples-of-ints like ((3,), (3,))
    shape: Tuple[int, ...]  # tuple of ints like (3, 3)
    dims: Optional[Tuple[str, ...]] = None  # None or tuple of names like ("x", "y")
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_json(cls, structure):
        if "fields" in structure["data_type"]:
            data_type = StructDtype.from_json(structure["data_type"])
        else:
            data_type = BuiltinDtype.from_json(structure["data_type"])
        dims = structure["dims"]
        if dims is not None:
            dims = tuple(dims)
        return cls(
            data_type=data_type,
            chunks=tuple(map(tuple, structure["chunks"])),
            shape=tuple(structure["shape"]),
            dims=dims,
            resizable=structure.get("resizable", False),
        )



import sqlite3
from contextlib import closing

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, record_history
from ..client.cache import Cache, CachedResponse
from ..server.app import build_app

tree = MapAdapter(
    {
        f"arr{i:03}": ArrayAdapter.from_array(i * numpy.arange(3), metadata={"i": i})
        for i in range(30)
    },
    metadata={"t": 1},
)


@pytest.fixture
def client():
    app = build_app(tree)
    with Context.from_app(app, cache=Cache()) as context:
        yield from_context(context)


def test_cache(client, tmpdir):
    cache = client.context.cache
    before_count = cache.count()
    before_size = cache.size()

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    after_count = cache.count()
    after_size = cache.size()
    assert after_count > before_count
    assert after_size > before_size

    # Second time: cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert isinstance(response, CachedResponse)


def test_no_cache(client):
    client.context.cache = None

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_lru_eviction(client):
    # First time: not cached
    client.context.cache.capacity = 5000
    num_items = len(client)
    for i in range(num_items):
        client.values()[i]

    # Not all the items fit.
    # If this fails, perhaps the bytesize of an item has drifted.
    # Tweak the capacity.
    assert client.context.cache.count() < num_items

    # Most recently accessed: still cached
    with record_history() as h:
        client.values()[i]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    # Least recently accessed: has been evicted
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        client.metadata
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_item_too_large_to_store(client):
    client.context.cache.max_item_size = 10

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: still not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_readonly_cache(client):
    # Start with a writable cache.

    # First time: not cached
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    orig_size = client.context.cache.size()

    # Now use the same file as readonly cache.
    filepath = client.context.cache.filepath
    ro_cache = client.context.cache = Cache(filepath, readonly=True)

    # Still cached (from before)
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    # Now look at something new...
    # First time: not cached
    with record_history() as h:
        client.values()[1]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: still not cached
    with record_history() as h:
        client.values()[1]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # And cache size has not changed
    assert ro_cache.size() == orig_size

    # Implementation detail: database connection is read-only,
    # for defense in depth.
    with pytest.raises(sqlite3.OperationalError):
        with closing(ro_cache._conn.cursor()) as cur:
            cur.execute("DELETE FROM responses")


def test_clear_cache(client):
    cache = client.context.cache
    client.values()[0]
    assert cache.size() > 0
    assert cache.count() > 0
    cache.clear()
    assert cache.size() == cache.count() == 0



import io

import h5py
import pandas
import pytest
import zarr

from ..catalog import in_memory
from ..catalog.register import register
from ..client import Context, from_context, tree
from ..server.app import build_app


@pytest.mark.asyncio
async def test_excel(tmpdir):
    df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    df.to_excel(tmpdir / "spreadsheet.xlsx")
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(catalog, tmpdir)
        tree(client)
        client["spreadsheet"]["Sheet1"].read()


@pytest.mark.asyncio
async def test_zarr_array(tmpdir):
    z = zarr.open(str(tmpdir / "za.zarr"), "w", shape=(3,), chunks=(3,), dtype="i4")
    z[:] = [1, 2, 3]
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(catalog, tmpdir)
        tree(client)
        client["za"].read()


@pytest.mark.asyncio
async def test_zarr_group(tmpdir):
    root = zarr.open(str(tmpdir / "zg.zarr"), "w")
    root.create_dataset("x", data=[1, 2, 3])
    root.create_dataset("y", data=[4, 5, 6])
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(catalog, tmpdir)
        tree(client)
        client["zg"].export(str(tmpdir / "stuff.h5"))
        client["zg"]["x"].read()
        client["zg"]["y"].read()


@pytest.mark.asyncio
async def test_hdf5(tmpdir):
    with h5py.File(str(tmpdir / "h.h5"), "w") as file:
        file["x"] = [1, 2, 3]
        group = file.create_group("g")
        group["y"] = [4, 5, 6]
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(catalog, tmpdir)
        tree(client)
        client["h"]["x"].read()
        client["h"]["g"]["y"].read()

        buffer = io.BytesIO()
        client.export(buffer, format="application/json")



import itertools

import numpy
import pandas
import sparse

from ..adapters.array import slice_and_shape_from_block_and_chunks
from ..structures.core import StructureFamily
from ..utils import path_from_uri


def load_block(uri):
    # TODO This can be done without pandas.
    # Better to use a plain I/O library.
    df = pandas.read_parquet(path_from_uri(uri))
    coords = df[df.columns[:-1]].values.T
    data = df["data"].values
    return coords, data


class SparseBlocksParquetAdapter:
    structure_family = StructureFamily.sparse

    def __init__(
        self,
        data_uris,
        structure,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        num_blocks = (range(len(n)) for n in structure.chunks)
        self.blocks = {}
        for block, uri in zip(itertools.product(*num_blocks), data_uris):
            self.blocks[block] = uri
        self._structure = structure
        self._metadata = metadata or {}
        self.specs = list(specs or [])
        self.access_policy = access_policy

    @classmethod
    def init_storage(
        cls,
        data_uri,
        structure,
    ):
        from ..server.schemas import Asset

        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)

        num_blocks = (range(len(n)) for n in structure.chunks)
        assets = [
            Asset(
                data_uri=f"{data_uri}/block-{'.'.join(map(str, block))}.parquet",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i, block in enumerate(itertools.product(*num_blocks))
        ]
        return assets

    def metadata(self):
        return self._metadata

    def write_block(self, data, block):
        uri = self.blocks[block]
        data.to_parquet(path_from_uri(uri))

    def write(self, data):
        if len(self.blocks) > 1:
            raise NotImplementedError
        uri = self.blocks[(0,) * len(self._structure.shape)]
        data.to_parquet(path_from_uri(uri))

    def read(self, slice=...):
        all_coords = []
        all_data = []
        for block, uri in self.blocks.items():
            coords, data = load_block(uri)
            offsets = []
            for b, c in zip(block, self.structure().chunks):
                offset = sum(c[:b])
                offsets.append(offset)
            global_coords = coords + [[i] for i in offsets]
            all_coords.append(global_coords)
            all_data.append(data)
        arr = sparse.COO(
            data=numpy.concatenate(all_data),
            coords=numpy.concatenate(all_coords, axis=-1),
            shape=self._structure.shape,
        )
        return arr[slice]

    def read_block(self, block, slice=...):
        coords, data = load_block(self.blocks[block])
        _, shape = slice_and_shape_from_block_and_chunks(block, self._structure.chunks)
        arr = sparse.COO(data=data[:], coords=coords[:], shape=shape)
        return arr[slice]

    def structure(self):
        return self._structure



"""
Use this examples like:

tiled serve pyobject --public tiled.examples.nexus:tree

To serve a different URL from the example hard-coded here, use the config:

```
# config.yml
authentication:
    allow_anonymous_access: true
trees:
    - path: /
      tree: tiled.examples.nexus:MapAdapter
      args:
          url: YOUR_URL_HERE
```

tiled serve config config.yml
"""
import io

import h5py
import httpx

from tiled.adapters.hdf5 import HDF5Adapter


def build_tree(url):
    # Download a Nexus file into a memory buffer.
    buffer = io.BytesIO(httpx.get(url, follow_redirects=True).content)
    # Access the buffer with h5py, which can treat it like a "file".
    file = h5py.File(buffer, "r")
    # Wrap the h5py.File in a MapAdapter to serve it with Tiled.
    return HDF5Adapter(file)


EXAMPLE_URL = "https://github.com/nexusformat/exampledata/blob/master/APS/EPICSareaDetector/hdf5/AgBehenate_228.hdf5?raw=true"  # noqa
tree = build_tree(EXAMPLE_URL)



import json
import os
import sqlite3
import threading
import typing as tp
from contextlib import closing
from datetime import datetime
from pathlib import Path

import appdirs
import httpx

from .utils import TiledResponse

CACHE_DATABASE_SCHEMA_VERSION = 1


class CachedResponse(TiledResponse):
    pass


def get_cache_key(request: httpx.Request) -> str:
    """Get the cache key from a request.

    The cache key is the str request url.

    Args:
        request: httpx.Request

    Returns:
        str: httpx.Request.url
    """
    return str(request.url)


def get_size(response, content=None):
    # get content or stream_content
    if hasattr(response, "_content"):
        size = len(response.content)
    elif content is None:
        raise httpx.ResponseNotRead()
    else:
        size = len(content)
    return size


def dump(response, content=None):
    # get content or stream_content
    if hasattr(response, "_content"):
        body = response.content
        is_stream = False
    elif content is None:
        raise httpx.ResponseNotRead()
    else:
        body = content
        is_stream = True

    return (
        response.status_code,
        json.dumps(response.headers.multi_items()),
        body,
        is_stream,
        response.encoding or None,
        len(body),
        datetime.now().timestamp(),
        0,  # never yet accessed
    )


def load(row, request=None):
    status_code, headers_json, body, is_stream, encoding = row
    headers = json.loads(headers_json)
    if is_stream:
        content = None
        stream = httpx.ByteStream(body)
    else:
        content = body
        stream = None
    response = CachedResponse(
        status_code, headers=headers, content=content, stream=stream
    )
    if encoding is not None:
        response.encoding = encoding

    if request is not None:
        response.request = request
    return response


def _create_tables(conn):
    with closing(conn.cursor()) as cur:
        cur.execute(
            """CREATE TABLE responses (
cache_key TEXT PRIMARY KEY,
status_code INTEGER,
headers JSON,
body BLOB,
is_stream INTEGER,
encoding TEXT,
size INTEGER,
time_created REAL,
time_last_accessed REAL
)"""
        )
        cur.execute("CREATE TABLE tiled_http_response_cache_version (version INTEGER)")
        cur.execute(
            "INSERT INTO tiled_http_response_cache_version (version) VALUES (?)",
            (CACHE_DATABASE_SCHEMA_VERSION,),
        )
        conn.commit()


def _prepare_database(filepath, readonly):
    Path(filepath).parent.mkdir(parents=True, exist_ok=True)
    if readonly:
        # The methods in Cache will not try to write when in readonly mode.
        # For extra safety we open a readonly connection to the database, so
        # that SQLite itself will prohibit writing.
        conn = sqlite3.connect(
            f"file:{filepath}?mode=ro", uri=True, check_same_thread=False
        )
    else:
        conn = sqlite3.connect(filepath, check_same_thread=False)
    cursor = conn.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = [row[0] for row in cursor.fetchall()]
    if not tables:
        # We have an empty database.
        _create_tables(conn)
    elif "tiled_http_response_cache_version" not in tables:
        # We have a nonempty database that we do not recognize.
        raise RuntimeError(
            f"Database at {filepath} is not empty and not recognized as a tiled HTTP response cache."
        )
    else:
        # We have a nonempty database that we recognize.
        cursor = conn.execute("SELECT * FROM tiled_http_response_cache_version;")
        (version,) = cursor.fetchone()
        if version != CACHE_DATABASE_SCHEMA_VERSION:
            # It is likely that this cache database will be very stable; we
            # *may* never need to change the schema. But if we do, we will
            # not bother with migrations. The cache is highly disposable.
            # Just silently blow it away and start over.
            Path(filepath).unlink()
            conn = sqlite3.connect(filepath, check_same_thread=False)
            _create_tables(conn)
    return conn


class Cache:
    def __init__(
        self,
        filepath=None,
        capacity=500_000_000,
        max_item_size=500_000,
        readonly=False,
    ):
        if filepath is None:
            # Resolve this here, not at module scope, because the test suite
            # injects TILED_CACHE_DIR env var to use a temporary directory.
            TILED_CACHE_DIR = Path(
                os.getenv("TILED_CACHE_DIR", appdirs.user_cache_dir("tiled"))
            )
            # TODO Detect filesystem of TILED_CACHE_DIR. If it is a networked filesystem
            # use a temporary database instead.
            filepath = TILED_CACHE_DIR / "http_response_cache.db"
        if capacity <= max_item_size:
            raise ValueError("capacity must be greater than max_item_size")
        self._capacity = capacity
        self._max_item_size = max_item_size
        self._readonly = readonly
        self._filepath = filepath
        self._owner_thread = threading.current_thread().ident
        self._conn = _prepare_database(filepath, readonly)

    def __repr__(self):
        return f"<{type(self).__name__} {str(self._filepath)!r}>"

    def write_safe(self):
        """
        Check that it is safe to write.

        SQLite is not threadsafe for concurrent _writes_.
        """
        return threading.current_thread().ident == self._owner_thread

    def __getstate__(self):
        return (self.filepath, self.capacity, self.max_item_size, self._readonly)

    def __setstate__(self, state):
        (filepath, capacity, max_item_size, readonly) = state
        self._capacity = capacity
        self._max_item_size = max_item_size
        self._readonly = readonly
        self._filepath = filepath
        self._owner_thread = threading.current_thread().ident
        self._conn = _prepare_database(filepath, readonly)

    @property
    def readonly(self):
        "If True, cache be read but not updated."
        return self._readonly

    @property
    def filepath(self):
        "Filepath of SQLite database storing cache data"
        return self._filepath

    @property
    def capacity(self):
        "Max capacity in bytes. Includes response bodies only."
        return self._capacity

    @capacity.setter
    def capacity(self, capacity):
        self._capacity = capacity

    @property
    def max_item_size(self):
        "Max size of a response body eligible for caching."
        return self._max_item_size

    @max_item_size.setter
    def max_item_size(self, max_item_size):
        self._max_item_size = max_item_size

    def clear(self):
        """
        Drop all entries from HTTP response cache.
        """
        if self.readonly:
            raise RuntimeError("Cannot clear read-only cache")
        if not self.write_safe():
            raise RuntimeError(
                "Cannot clear cache from a different thread than the one it was created on"
            )
        with closing(self._conn.cursor()) as cur:
            cur.execute("DELETE FROM responses")
            self._conn.commit()

    def get(self, request: httpx.Request) -> tp.Optional[httpx.Response]:
        """Get cached response from Cache.

        We use the httpx.Request.url as key.

        Args:
            request: httpx.Request

        Returns:
            tp.Optional[httpx.Response]
        """
        with closing(self._conn.cursor()) as cur:
            cache_key = get_cache_key(request)
            row = cur.execute(
                """SELECT
status_code, headers, body, is_stream, encoding
FROM
responses
WHERE cache_key = ?""",
                (cache_key,),
            ).fetchone()
            if row is None:
                return None
            if (not self.readonly) and self.write_safe():
                cur.execute(
                    """UPDATE responses
    SET time_last_accessed = ?
    WHERE cache_key = ?""",
                    (datetime.now().timestamp(), cache_key),
                )
            self._conn.commit()

        return load(row, request)

    def set(
        self,
        *,
        request: httpx.Request,
        response: httpx.Response,
        content: tp.Optional[bytes] = None,
    ) -> None:
        """Set new response entry in cache.

        In case the response does not yet have a '_content' property, content should
        be provided in the optional 'content' kwarg (usually using a callback)

        Parameters
        ----------
        request: httpx.Request
        response: httpx.Response, to cache
        content (bytes, optional): Defaults to None, should be provided in case
            response that not have yet content.
        """
        if self.readonly:
            raise RuntimeError("Cache is readonly")
        if not self.write_safe():
            raise RuntimeError("Write is not safe from another thread")
        incoming_size = get_size(response, content)
        if incoming_size > self.max_item_size:
            # Decline to store.
            return False
        with closing(self._conn.cursor()) as cur:
            (total_size,) = cur.execute("SELECT SUM(size) FROM responses").fetchone()
            total_size = total_size or 0  # If empty, total_size is None.
            while (incoming_size + total_size) > self.capacity:
                # Cull to make space.
                (cache_key, size) = cur.execute(
                    """SELECT
cache_key, size
FROM responses
ORDER BY time_last_accessed ASC"""
                ).fetchone()
                cur.execute("DELETE FROM responses WHERE cache_key = ?", (cache_key,))
                total_size -= size
            cur.execute(
                """INSERT OR REPLACE INTO responses
(cache_key, status_code, headers, body, is_stream, encoding, size, time_created, time_last_accessed)
VALUES
(?, ?, ?, ?, ?, ?, ?, ?, ?)""",
                (get_cache_key(request),) + dump(response, content),
            )
            self._conn.commit()
        return True

    def delete(self, request: httpx.Request) -> None:
        """Delete an entry from cache.

        Args:
            request: httpx.Request
        """
        if self.readonly:
            raise RuntimeError("Cache is readonly")
        if not self.write_safe():
            raise RuntimeError("Write is not safe from another thread")
        with closing(self._conn.cursor()) as cur:
            cur.execute(
                "DELETE FROM responses WHERE cache_key=?", (get_cache_key(request),)
            )
            self._conn.commit()

    def size(self):
        "Size of response bodies in bytes (does not count headers and other auxiliary info)"
        with closing(self._conn.cursor()) as cur:
            (total_size,) = cur.execute("SELECT SUM(size) FROM responses").fetchone()
        return total_size or 0  # If emtpy, total_size is None.

    def count(self):
        "Number of responses cached"
        with closing(self._conn.cursor()) as cur:
            (count,) = cur.execute("SELECT COUNT(*) FROM responses").fetchone()
        return count or 0  # If empty, count is None.

    def close(self) -> None:
        """Close cache."""
        self._conn.close()



import importlib

from ..serialization.table import XLSX_MIME_TYPE
from ..utils import OneShotCachedMap

# This maps MIME types (i.e. file formats) for appropriate Readers.
# OneShotCachedMap is used to defer imports. We don't want to pay up front
# for importing Readers that we will not actually use.
PARQUET_MIMETYPE = "application/x-parquet"
SPARSE_BLOCKS_PARQUET_MIMETYPE = "application/x-parquet;structure=sparse"
ZARR_MIMETYPE = "application/x-zarr"
AWKWARD_BUFFERS_MIMETYPE = "application/x-awkward-buffers"
DEFAULT_ADAPTERS_BY_MIMETYPE = OneShotCachedMap(
    {
        "image/tiff": lambda: importlib.import_module(
            "...adapters.tiff", __name__
        ).TiffAdapter,
        "multipart/related;type=image/tiff": lambda: importlib.import_module(
            "...adapters.tiff", __name__
        ).TiffSequenceAdapter.from_uris,
        "text/csv": lambda: importlib.import_module(
            "...adapters.csv", __name__
        ).read_csv,
        XLSX_MIME_TYPE: lambda: importlib.import_module(
            "...adapters.excel", __name__
        ).ExcelAdapter.from_uri,
        "application/x-hdf5": lambda: importlib.import_module(
            "...adapters.hdf5", __name__
        ).HDF5Adapter.from_uri,
        "application/x-netcdf": lambda: importlib.import_module(
            "...adapters.netcdf", __name__
        ).read_netcdf,
        PARQUET_MIMETYPE: lambda: importlib.import_module(
            "...adapters.parquet", __name__
        ).ParquetDatasetAdapter,
        SPARSE_BLOCKS_PARQUET_MIMETYPE: lambda: importlib.import_module(
            "...adapters.sparse_blocks_parquet", __name__
        ).SparseBlocksParquetAdapter,
        ZARR_MIMETYPE: lambda: importlib.import_module(
            "...adapters.zarr", __name__
        ).read_zarr,
        AWKWARD_BUFFERS_MIMETYPE: lambda: importlib.import_module(
            "...adapters.awkward_buffers", __name__
        ).AwkwardBuffersAdapter.from_directory,
    }
)

# We can mostly rely on mimetypes.types_map for the common ones
# ('.csv' -> 'text/csv', etc.) but we supplement here for some
# of the more exotic ones that not all platforms know about.
DEFAULT_MIMETYPES_BY_FILE_EXT = {
    # This is the "official" file extension.
    ".h5": "application/x-hdf5",
    # This is NeXus. We may want to invent a special media type
    # like 'application/x-nexus' for this, but I'll punt that for now.
    # Needs thought about how to encode the various types of NeXus
    # (media type arguments, for example).
    ".nxs": "application/x-hdf5",
    # These are unofficial but common file extensions.
    ".hdf": "application/x-hdf5",
    ".hdf5": "application/x-hdf5",
    # on opensuse csv -> text/x-comma-separated-values
    ".csv": "text/csv",
    ".zarr": ZARR_MIMETYPE,
}
